"""
Class for reading and manipulating head-related transfer functions. Reads files in .sofa format (started before
python implementations of the sofa conventions were available -> will be migrated to use pysofaconventions!)
"""

import copy
import warnings
import pathlib
import pickle
import bz2
import numpy
import datetime
import itertools
try:
    import matplotlib
    from matplotlib import pyplot as plt
except ImportError:
    matplotlib, plt = False, False
try:
    from mpl_toolkits.mplot3d import Axes3D
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    Axes3D = False
    make_axes_locatable = False
try:
    import h5netcdf
except ImportError:
    h5netcdf = False
try:
    import netCDF4
except ImportError:
    netCDF4 = False
try:
    import scipy.signal
    import scipy.spatial
except ImportError:
    scipy = False
from slab.signal import Signal
from slab.filter import Filter
from slab.sound import Sound
from collections import namedtuple

_kemar = None

class HRTF:
    """
    Class for reading and manipulating head-related transfer functions with attributes and functions to manage them.

    Arguments:
        data (str | Filter | numpy.ndarray): Typically, this is the path to a file in the .sofa format.
            The file is then loaded and the data of each source for which the transfer function was recorded is stored
            as a Filter object in the `data` attribute. Instead of a file name, the data can be passed directly as
            Filter or numpy array. Given a `Filter`, every filter channel in the instance is taken as a source (this
            does not result in a typical HRTF object and is only intended for equalization filter banks). Given a 3D
            array, the first dimension represents the sources, the second the number of taps per filter and the last the
            number of filter channels per filter (should be always 2, for left and right ear).
        datatype (None | string): type of the HRTF filter bank, can be 'FIR' for finite impulse response filters or 'TF'
            for Fourier filters.
        samplerate (None | float): rate at which the data was acquired, only relevant when not loading from .sofa file
        sources (None | array): positions of the recorded sources, only relevant when not loading from .sofa file
        listener (None | list | dict): position of the listener, only relevant when not loading from .sofa file
        verbose (bool): print out items when loading .sofa files, defaults to False

    Attributes:
        .data (list): The HRTF data. The elements of the list are instances of slab.Filter.
        .datatype (string): Type of the HRTF filter bank.
        .samplerate (int): sampling rate at which the HRTF data was acquired.
        .sources (named tuple): Cartesian coordinates (x, y, z), vertical-polar and interaural-polar coordinates
            (azimuth, elevation, distance) of all sources.
        .n_sources (int): The number of sources in the HRTF.
        .n_elevations (int): The number of elevations in the HRTF.
        .listener (dict): A dictionary containing the position of the listener ("pos"), the point which the listener
            is fixating ("view"), the point 90° above the listener ("up") and vectors from the listener to those points.
    Example:
        import slab
        hrtf = slab.HRTF.kemar() # use inbuilt KEMAR data
        sourceidx = hrtf.cone_sources(20)
        hrtf.plot_sources(sourceidx)
        hrtf.plot_tf(sourceidx, ear='left')
    """
    # instance properties
    n_sources = property(fget=lambda self: len(self.sources.vertical_polar),
                         doc='The number of sources in the HRTF.')
    n_elevations = property(fget=lambda self: len(self.elevations()),
                            doc='The number of elevations in the HRTF.')

    def __init__(self, data, datatype=None, samplerate=None, sources=None, listener=None, verbose=False):
        if isinstance(data, pathlib.Path):
            data = str(data)
        if isinstance(data, str):
            if samplerate is not None:
                raise ValueError('Cannot specify samplerate when initialising HRTF from a file.')
            if pathlib.Path(data).suffix != '.sofa':
                raise NotImplementedError('Only .sofa files can be read.')
            f = HRTF._sofa_load(data, verbose)
            self.data, self.datatype, self.samplerate = HRTF._sofa_get_data(f)
            sources, coordinate_system = HRTF._sofa_get_sources(f)
            self.sources = HRTF._get_coordinates(sources, coordinate_system)
            self.listener = HRTF._sofa_get_listener(f)
        elif isinstance(data, Filter):
            # This is a hacky shortcut for casting a filterbank as HRTF. Avoid unless you know what you are doing.
            if sources is None:
                raise ValueError('Must provide spherical source positions when initializing HRTF from a Filter object.')
            self.samplerate = data.samplerate
            fir = data.fir
            if 'IR' in fir:
                self.datatype = 'FIR'
            else:
                self.datatype = 'TF'
            # reshape the filterbank data to fit into HRTF (ind x taps x ear)
            data = data.data.T[..., None]
            self.data = []
            for idx in range(data.shape[0]):
                self.data.append(Filter(data[idx, :, :].T, self.samplerate, fir=fir))
            self.sources = HRTF._get_coordinates(sources, 'spherical')
            if listener is None:
                self.listener = [0, 0, 0]
            else:
                self.listener = listener
        elif isinstance(data, numpy.ndarray):
            if samplerate is None:  # should have shape (ind x ear x taps), 2 x n_taps filter (left right)
                raise ValueError('Must specify samplerate when initialising HRTF from an array.')
            self.samplerate = samplerate
            if datatype is None or type(datatype) is not str:
                raise ValueError('Must specify data type (FIR or TF) when initialising HRTF from an array.')
            self.datatype = datatype.upper()
            if self.datatype == 'FIR':
                fir = 'IR'
            elif self.datatype == 'TF':
                fir = 'TF'
            else:
                raise ValueError(f'Unsupported data type: {datatype}')
            if sources is None:
                raise ValueError('Must provide vertical-polar source positions when initializing HRTF from an array.')
            self.sources = HRTF._get_coordinates(sources, 'spherical')
            self.data = []
            for idx in range(data.shape[0]):
                self.data.append(Filter(data[idx, :, :].T, self.samplerate, fir=fir))
            if listener is None:
                self.listener = [0, 0, 0]
            else:
                self.listener = listener
        else:
            raise ValueError(f'Unsupported data type: {type(data)}')

    def __repr__(self):
        return f'{type(self)} (\n{repr(self.data)} \n{repr(self.samplerate)})'

    def __str__(self):
        return f'{type(self)}, datatype {self.datatype}, sources {self.n_sources}, ' \
               f'elevations {self.n_elevations}, samplerate {self.samplerate}, samples {self[0].n_samples}'

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __setitem__(self, key, value):
        return self.data.__setitem__(key, value)

    # Static methods (used in __init__)
    @staticmethod
    def _sofa_load(filename, verbose=False):
        """
        Read a SOFA file.

        Arguments:
            f (h5netcdf.core.File): data as returned by the `_sofa_load` method.
        Returns:
            (h5netcdf.core.File): the data from the .sofa file.
        """
        if h5netcdf is False:
            raise ImportError('Reading from sofa files requires h5py and h5netcdf.')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            f = h5netcdf.File(filename, 'r')
        if verbose:
            f.items()
        return f

    @staticmethod
    def _sofa_get_data(f):
        """
        Read the impulse response or transfer functions from the SOFA data and store them in a Filter object.
        Return the Filter object in the `data` attribute, the data type and samplerate.

        Arguments:
            f (h5netcdf.core.File): data as returned by the `_sofa_load` method.
        Returns:
            data (list): a list of Filter objects containing the HRTF data.
            datatype (string): the data type used in the SOFA file
            samplerate (float): the sampling rate in Hz.
        """
        samplerate = HRTF._sofa_get_samplerate(f)
        datatype = f.attrs['DataType']
        data = []
        if datatype == 'FIR':
            ir_data = numpy.array(f.variables['Data.IR'], dtype='float')
            for idx in range(ir_data.shape[0]):
                data.append(Filter(ir_data[idx, :, :].T, samplerate, fir='IR'))  # n_taps x 2 (left, right) filter
        elif datatype == 'TF':
            data_real = numpy.array(f.variables['Data.Real'], dtype='float')
            data_imag = numpy.array(f.variables['Data.Imag'], dtype='float')
            tf_data = numpy.abs(numpy.vectorize(complex)(data_real, data_imag))
            for idx in range(tf_data.shape[0]):
                data.append(Filter(tf_data[idx, :, :].T, samplerate, fir='TF'))
        else:
            raise NotImplementedError('Unsuppored datatype: {self.datatype}')
        return data, datatype, samplerate

    @staticmethod
    def _sofa_get_samplerate(f):
        """
        Returns the sampling rate of the recordings. If the sampling rate is not given in Hz, the function assumes
        it is given in kHz and multiplies by 1000 to convert to Hz.

        Arguments:
            f (h5netcdf.core.File): data as returned by the `_sofa_load` method.
        Returns:
            (float): the sampling rate in Hz.
        """
        attr = dict(f.variables['Data.SamplingRate'].attrs.items())  # get attributes as dict
        unit = attr['Units']  # extract and decode Units
        if unit in ('hertz', 'Hz'):
            return (numpy.array(f.variables['Data.SamplingRate'], dtype='int'))
        warnings.warn('Unit other than Hz. ' + unit + '. Assuming kHz.')
        return 1000 * (numpy.array(f.variables['Data.SamplingRate'], dtype='int'))

    @staticmethod
    def _sofa_get_sources(f):
        """
        Returns an array of positions of all sound sources.

        Arguments:
            f (h5netcdf.core.File): data as returned by the _sofa_load method.
        Returns:
            (numpy.ndarray): coordinates of all sources.
            (string): coordinate system used in the SOFA file.
        """
        # spherical coordinates, (azi,ele,radius), azi 0..360 (0=front, 90=left, 180=back), ele -90..90
        sources = numpy.array(f.variables['SourcePosition'], dtype='float')
        attr = dict(f.variables['SourcePosition'].attrs.items())  # get attributes as dict
        coordinate_system = attr['Type'].split(',')[0]  # extract and decode Units
        return sources, coordinate_system

    @staticmethod
    def _sofa_get_listener(f):
        """
        Returns dict with listeners positional information - used for plotting.

        Arguments:
            f (h5netcdf.core.File): data as returned by the `_sofa_load()` method.
        Returns:
            (dict): position of the listener ("pos"), the point which the listener is fixating ("view")
                the point 90° above the listener ("up") and vectors from the listener to those points.
        """
        lis = {'pos': numpy.array(f.variables['ListenerPosition'], dtype='float')[0],
               'view': numpy.array(f.variables['ListenerView'], dtype='float')[0],
               'up': numpy.array(f.variables['ListenerUp'], dtype='float')[0]}
        lis['viewvec'] = numpy.concatenate([lis['pos'], lis['pos']+lis['view']])
        lis['upvec'] = numpy.concatenate([lis['pos'], lis['pos']+lis['up']])
        return lis

    @staticmethod
    def _get_coordinates(sources, coordinate_system):
        """
        Returns the sound source positions in three different coordinate systems:
        cartesian, vertical-polar and interaural-polar.

        Arguments:
            sources (numpy.ndarray): sound source coordinates in cartesian coordinates (x, y, z),
                vertical-polar or interaural-polar coordinates (azimuth, elevation, distance).
            coordinate_system (string): type of the provided coordinates. Can be 'cartesian',
                'vertical_polar' or 'interaural_polar'.
        Returns:
            (named tuple): cartesian, vertical-polar and interaural-polar coordinates of all sources.
        """
        if isinstance(sources, (list, tuple)):
            sources = numpy.array(sources)
        if len(sources.shape) == 1:  # a single location (vector) needs to be converted to a 2d matrix
            sources = sources[numpy.newaxis, ...]
        sources = sources.astype('float64')
        source_coordinates = namedtuple('sources', 'cartesian vertical_polar interaural_polar')
        if coordinate_system == 'spherical':
            vertical_polar = sources
            cartesian = HRTF._vertical_polar_to_cartesian(vertical_polar)
            interaural_polar = HRTF._vertical_polar_to_interaural_polar(vertical_polar)
        elif coordinate_system == 'interaural':
            interaural_polar = sources
            cartesian = HRTF._interaural_polar_to_cartesian(interaural_polar)
            vertical_polar = HRTF._cartesian_to_vertical_polar(cartesian)
        elif coordinate_system == 'cartesian':
            cartesian = sources
            vertical_polar = HRTF._cartesian_to_vertical_polar(cartesian)
            interaural_polar = HRTF._vertical_polar_to_interaural_polar(vertical_polar)
        else:
            warnings.warn('Unrecognized coordinate system for source positions: ' + coordinate_system)
            return None
        return source_coordinates(cartesian.astype('float16'), vertical_polar.astype('float16'),
                                  interaural_polar.astype('float16'))

    @staticmethod
    def _vertical_polar_to_cartesian(vertical_polar):
        """
        Convert vertical-polar to cartesian coordinates.

        Arguments:
            vertical_polar (numpy.ndarray): vertical-polar coordinates (azimuth, elevation, distance).
        Returns:
            (numpy.ndarray): cartesian coordinates.
        """
        cartesian = numpy.zeros_like(vertical_polar)
        azimuths = numpy.deg2rad(vertical_polar[:, 0])
        elevations = numpy.deg2rad(90 - vertical_polar[:, 1])
        r = vertical_polar[:, 2].mean()  # get radii of sound sources
        cartesian[:, 0] = r * numpy.cos(azimuths) * numpy.sin(elevations)
        cartesian[:, 1] = r * numpy.sin(elevations) * numpy.sin(azimuths)
        cartesian[:, 2] = r * numpy.cos(elevations)
        return cartesian

    @staticmethod
    def _interaural_polar_to_cartesian(interaural_polar):
        """
        Convert interaural-polar to cartesian coordinates.

        Arguments:
            interaural_polar (numpy.ndarray): interaural-polar coordinates (azimuth, elevation, distance).
        Returns:
            (numpy.ndarray): cartesian coordinates.
        """
        cartesian = numpy.zeros_like(interaural_polar)
        azimuths = numpy.deg2rad(interaural_polar[:, 0])
        elevations = numpy.deg2rad(90 - interaural_polar[:, 1])
        r = interaural_polar[:, 2].mean()  # get radii of sound sources
        cartesian[:, 0] = r * numpy.cos(azimuths) * numpy.sin(elevations)
        cartesian[:, 1] = r * numpy.sin(azimuths)
        cartesian[:, 2] = r * numpy.cos(elevations) * numpy.cos(azimuths)
        return cartesian

    @staticmethod
    def _cartesian_to_vertical_polar(cartesian):
        """
        Convert cartesian to vertical-polar coordinates.

        Arguments:
            cartesian (numpy.ndarray): cartesian coordinates (azimuth, elevation, distance).
        Returns:
            (numpy.ndarray): vertical-polar coordinates.
        """
        vertical_polar = numpy.zeros_like(cartesian)
        xy = cartesian[:, 0] ** 2 + cartesian[:, 1] ** 2
        vertical_polar[:, 0] = numpy.rad2deg(numpy.arctan2(cartesian[:, 1], cartesian[:, 0]))
        vertical_polar[vertical_polar[:, 0] < 0, 0] += 360
        vertical_polar[:, 1] = 90 - numpy.rad2deg(numpy.arctan2(numpy.sqrt(xy), cartesian[:, 2]))
        vertical_polar[:, 2] = numpy.sqrt(xy + cartesian[:, 2] ** 2)
        return vertical_polar

    @staticmethod
    def _vertical_polar_to_interaural_polar(vertical_polar):
        """
        Convert vertical-polar to interaural-polar coordinates.

        Arguments:
            vertical_polar (numpy.ndarray): cartesian coordinates (azimuth, elevation, distance).
        Returns:
            (numpy.ndarray): interaural-polar coordinates.
        """
        interaural_polar = numpy.zeros_like(vertical_polar)
        azimuths = numpy.deg2rad(vertical_polar[:, 0])
        elevations = numpy.deg2rad(vertical_polar[:, 1])
        interaural_polar[:, 0] = numpy.rad2deg(numpy.arcsin(numpy.cos(elevations) * numpy.sin(azimuths)))
        with numpy.errstate(divide='ignore'):
            interaural_polar[:, 1] = (numpy.pi / 2) - numpy.arctan(((1 / numpy.tan(elevations)) * numpy.cos(azimuths)))
        interaural_polar[elevations < 0, 1] += numpy.pi
        interaural_polar[:, 1] = numpy.rad2deg(interaural_polar[:, 1])
        interaural_polar[:, 2] = vertical_polar[:, 2]
        return interaural_polar

    def apply(self, source, sound, allow_resampling=True):
        """
        Apply a filter from the HRTF set to a sound. The sound will be recast as slab.Binaural. If the samplerates
        of the sound and the HRTF are unequal and `allow_resampling` is True, then the sound will be resampled to the
        filter rate, filtered, and then resampled to the original rate.
        The filtering is done with `scipy.signal.fftconvolve`.

        Arguments:
            source (int): the source index of the binaural filter in self.data.
            sound (slab.Signal | slab.Sound | slab.Binaural): the sound to be rendered spatially.
        Returns:
            (slab.Binaural): a spatialized copy of `sound`.
        """
        from slab.binaural import Binaural  # importing here to avoid circular import at top of class
        if self.datatype == 'FIR':
            if (sound.samplerate != self.samplerate) and (not allow_resampling):
                raise ValueError('Filter and sound must have same sampling rates.')
            original_rate = sound.samplerate
            sound = sound.resample(self.samplerate)  # does nothing if samplerates are the same
            out = self[source].apply(sound)
            return Binaural(out.resample(original_rate))
        if self.datatype == 'TF':  # Filter.apply DTF as Fourier filter
            return self[source].apply(sound)

    def elevations(self):
        """
        Get all different elevations at which sources where recorded . Note: This currently only works as
        intended for HRTFs recorded in horizontal rings.

        Returns:
             (list): a sorted list of source elevations.
        """
        return sorted(list(set(numpy.round(self.sources.vertical_polar[:, 1]))))

    def plot_tf(self, sourceidx, ear='left', xlim=(1000, 18000), n_bins=None, kind='waterfall',
                linesep=20, xscale='linear', show=True, axis=None):
        """
        Plot transfer functions at a list of source indices.

        Arguments:
            ear (str): the ear from which data is plotted. Can be 'left', 'right', or 'both'.
            sourceidx (list of int): sources to plot. Typically be generated using the `hrtf.cone_sources` Method.
            xlim (tuple of int): frequency range of the plot
            n_bins (int) : passed to :meth:`slab.Filter.tf` and determines frequency resolution
            kind (str): type of plot to draw. Can be `waterfall` (as in Wightman and Kistler, 1989),
                `image` (as in Hofman, 1998) or 'surface' (as in Schnupp and Nelken, 2011).
            linesep (int): vertical distance between transfer functions in the waterfall plot
            xscale (str): sets x-axis scaling ('linear', 'log')
            show (bool): If True, show the plot immediately
            axis (matplotlib.axes._subplots.AxesSubplot): Axis to draw the plot on
        """
        if matplotlib is False:
            raise ImportError('Plotting HRTFs requires matplotlib.')
        if ear == 'left':
            chan = 0
        elif ear == 'right':
            chan = 1
        elif ear == 'both':
            if axis is not None and not isinstance(axis, (list, numpy.ndarray)):
                raise ValueError("Axis must be a list of length two when plotting left and right ear!")
            elif axis is None:
                axis = [None, None]
            if kind == 'image':
                fig1 = self.plot_tf(sourceidx, ear='left', xlim=xlim, axis=axis[0], show=show,
                                    linesep=linesep, n_bins=n_bins, kind='image', xscale=xscale)
                fig2 = self.plot_tf(sourceidx, ear='right', xlim=xlim, axis=axis[1], show=show,
                                    linesep=linesep, n_bins=n_bins, kind='image', xscale=xscale)
            elif kind == "waterfall":
                fig1 = self.plot_tf(sourceidx, ear='left', xlim=xlim, axis=axis[0], show=show,
                                    linesep=linesep, n_bins=n_bins, kind='waterfall', xscale=xscale)
                fig2 = self.plot_tf(sourceidx, ear='right', xlim=xlim, axis=axis[1], show=show,
                                    linesep=linesep, n_bins=n_bins, kind='waterfall', xscale=xscale)
            else:
                raise ValueError("'Kind' must be either 'waterfall' or 'image'!")
            return fig1, fig2
        else:
            raise ValueError("Unknown value for ear. Use 'left', 'right', or 'both'")
        if not axis:
            fig, axis = plt.subplots()
        else:
            fig = axis.figure
        if kind == 'waterfall':
            vlines = numpy.arange(0, len(sourceidx)) * linesep
            for idx, source in enumerate(sourceidx):
                filt = self[source]
                freqs, h = filt.tf(channels=chan, n_bins=n_bins, show=False)
                axis.plot(freqs, h + vlines[idx],
                          linewidth=0.75, color='0.0', alpha=0.7)
            ticks = vlines[::2]  # plot every second elevation
            labels = numpy.round(self.sources.vertical_polar[sourceidx, 1]*2, decimals=-1)/2
            # plot every third elevation label, omit comma to save space
            labels = labels[::2].astype(int)
            axis.set(yticks=ticks, yticklabels=labels)
            axis.grid(visible=True, axis='y', which='both', linewidth=0.25)
            axis.plot([xlim[0]+500, xlim[0]+500], [vlines[-1]+10, vlines[-1] +
                      10+linesep], linewidth=1, color='0.0', alpha=0.9)
            axis.text(x=xlim[0]+600, y=vlines[-1]+10+linesep/2,
                      s=str(linesep)+'dB', va='center', ha='left', fontsize=6, alpha=0.7)
        elif kind == 'image' or 'surface':
            if not n_bins:
                img = numpy.zeros((self[sourceidx[0]].n_taps, len(sourceidx)))
            else:
                img = numpy.zeros((n_bins, len(sourceidx)))
            elevations = self.sources.vertical_polar[sourceidx, 1]
            for idx, source in enumerate(sourceidx):
                filt = self[source]
                freqs, h = filt.tf(channels=chan, n_bins=n_bins, show=False)
                img[:, idx] = h.flatten()
            img[img < -25] = -25  # clip at -40 dB transfer
            if kind == 'image':
                contour = axis.contourf(freqs[freqs <= xlim[1]], elevations, img.T[:, freqs <= xlim[1]], cmap='hot', origin='upper', levels=20)
                divider = make_axes_locatable(axis)
                cax = divider.append_axes('right', size='5%', pad=0.05)
                fig.colorbar(contour, cax, orientation="vertical")
            elif kind == 'surface':
                xi, yi = numpy.meshgrid(freqs, elevations)  # interpolate to smooth surface plot
                spline = scipy.interpolate.Rbf(xi, yi, img.T, function='thin_plate')  # interpolator instance
                x, y = numpy.meshgrid(numpy.linspace(freqs.min(), freqs.max(), len(freqs)),
                                      numpy.linspace(elevations.min(), elevations.max(), 100))
                z = spline(x, y)
                x[x < xlim[0]] = numpy.nan  # trim edges
                x[x > xlim[1]] = numpy.nan
                fig, axis = plt.subplots()
                axis.axis('off')
                axis = plt.axes(projection='3d')
                contour = axis.plot_surface(x, y, z, rcount=200, ccount=200, cmap='cool')
                fig.colorbar(contour, fraction=0.046, pad=0.04, orientation="horizontal")
        else:
            raise ValueError("Unknown plot type. Use 'waterfall' or 'image'.")
        axis.xaxis.set_major_formatter(
            matplotlib.ticker.FuncFormatter(lambda x, pos: str(int(x / 1000))))
        axis.autoscale(tight=True)
        axis.tick_params('both', length=2, pad=2)
        if kind == 'surface':
            axis.set(xlabel='Frequency [Hz]', ylabel='Elevation [˚]', zlabel='Pinna gain [dB]',
                     xlim=xlim, xscale=xscale)
        else:
            axis.set(xlabel='Frequency [Hz]', ylabel='Elevation [˚]', xlim=xlim,
                     xscale=xscale)
        if show:
            plt.show()

    def diffuse_field_avg(self):
        """
        Compute the diffuse field average transfer function, i.e. the constant non-spatial portion of a set of HRTFs.
        The filters for all sources are averaged, which yields an unbiased average only if the sources are uniformly
        distributed around the head.

        Returns:
             (Filter): the diffuse field average as FFR filter object.
        """
        dfa = []
        for source in range(self.n_sources):
            filt = self[source]
            for chan in range(filt.n_channels):
                _, h = filt.tf(channels=chan, show=False)
                dfa.append(h)
        dfa = 10 ** (numpy.mean(dfa, axis=0)/20)  # average and convert from dB to gain
        return Filter(dfa, fir='IR', samplerate=self.samplerate)

    def diffuse_field_equalization(self, dfa=None):
        """
        Equalize the HRTF by dividing each filter by the diffuse field average. The resulting filters have a mean
        close to 0 and are Fourier filters.

        Arguments:
            dfa (None): Filter object containing the diffuse field average transfer function of the HRTF.
                If none is provided, the `diffuse_field_avg` method is called to obtain it.
        Returns:
            (HRTF): diffuse field equalized version of the HRTF.
        """
        if dfa is None:
            dfa = self.diffuse_field_avg()
        # invert the diffuse field average
        dfa.data = 1/dfa.data
        dtfs = copy.deepcopy(self)
        # apply the inverted filter to the HRTFs
        for source in range(dtfs.n_sources):
            filt = dtfs.data[source]
            _, h = filt.tf(show=False)
            h = 10 ** (h / 20) * dfa
            dtfs.data[source] = Filter(data=h, fir='TF', samplerate=self.samplerate)
        return dtfs

    def cone_sources(self, cone=0, full_cone=False):
        """
        Get all sources of the HRTF that lie on a "cone of confusion". The cone is a vertical off-axis sphere
        slice. All sources that lie on the cone have the same interaural level and time difference.
        Note: This currently only works as intended for HRTFs recorded in horizontal rings.

        Arguments:
            cone (int | float): azimuth of the cone center in degree.
            full_cone (bool): If True, return all sources that lie on the cone, otherwise, return only sources
                in front of the listener.
        Returns:
            (list): elements of the list are the indices of sound sources on the frontal half of the cone.
        Examples::

            import HRTF
            hrtf = slab.HRTF.kemar()
            sourceidx = hrtf.cone_sources(20)  # get the source indices
            print(hrtf.sources[sourceidx])  # print the coordinates of the source indices
            hrtf.plot_sources(sourceidx)  # show the sources in a 3D plot
        """
        cone = numpy.sin(numpy.deg2rad(cone))
        eles = self.elevations()
        _cartesian = self.sources.cartesian / 1.4  # get cartesian coordinates on the unit sphere
        out = []
        for ele in eles:  # for each elevation, find the source closest to the reference y
            if full_cone == False:  # only return cone sources in front of listener
                subidx, = numpy.where((numpy.round(self.sources.vertical_polar[:, 1]) == ele) & (_cartesian[:, 0] >= 0))
            else:  # include cone sources behind listener
                subidx, = numpy.where(numpy.round(self.sources.vertical_polar[:, 1]) == ele)
            cmin = numpy.min(numpy.abs(_cartesian[subidx, 1] - cone).astype('float16'))
            if cmin < 0.05:  # only include elevation where the closest source is less than 5 cm away
                idx, = numpy.where((numpy.round(self.sources.vertical_polar[:, 1]) == ele) & (
                        numpy.abs(_cartesian[:, 1] - cone).astype('float16') == cmin))  # avoid rounding error
                out.append(idx[0])
                if full_cone and len(idx) > 1:
                    out.append(idx[1])
        return sorted(out, key=lambda x: self.sources.vertical_polar[x, 1])

    def elevation_sources(self, elevation=0):
        """
        Get the indices of sources along a horizontal sphere slice at the given `elevation`.

        Arguments:
            elevation (int | float): The elevation of the sources in degree. The default returns sources along
                the frontal horizon.
        Returns:
            (list): indices of the sound sources. If the hrtf does not contain the specified `elevation` an empty
                list is returned.
        """
        idx = numpy.where((self.sources.vertical_polar[:, 1] == elevation) & (
                (self.sources.vertical_polar[:, 0] <= 90) | (self.sources.vertical_polar[:, 0] >= 270)))
        return idx[0].tolist()

    def tfs_from_sources(self, sources, n_bins=96):
        """
        Get the transfer function from sources in the hrtf.

        Arguments:
            sources (list): Indices of the sources (as generated for instance with the `HRTF.cone_sources` method),
                for which the transfer function is extracted.
            n_bins (int): The number of frequency bins for each transfer function.
        Returns:
            (numpy.ndarray): 2-dimensional array where the first dimension represents the frequency bins and the
                second dimension represents the sources.
        """
        n_sources = len(sources)
        tfs = numpy.zeros((n_bins, n_sources))
        for idx, source in enumerate(sources):
            _, jwd = self[source].tf(channels=0, n_bins=n_bins, show=False)
            tfs[:, idx] = jwd.flatten()
        return tfs

    def interpolate(self, azimuth=0, elevation=0, method='nearest', plot_tri=False):
        """
        Interpolate a filter at a given azimuth and elevation from the neighboring HRTFs. A weighted average of the
        3 closest HRTFs in the set is computed in the spectral domain with barycentric weights. The resulting filter
        values vary smoothly with changes in azimuth and elevation. The fidelity of the interpolated filter decreases
        with increasing distance of the closest sources and should only be regarded as appropriate approximation when
        the contributing filters are less than 20˚ away.

        Arguments:
            azimuth (float): the azimuth component of the direction of the interpolated filter
            elevation (float): the elevation component of the direction of the interpolated filter
            method (str): interpolation method, 'nearest' returns the filter of the nearest direction. Any other string
                returns a barycentric interpolation.
            plot_tri (bool): plot the triangulation of source positions used of interpolation. Useful for checking
                for areas where the interpolation may not be accurate (look for irregular or elongated triangles).
        Returns:
            (slab.Filter): an IR-type binaural Filter
        """
        from slab.binaural import Binaural  # importing here to avoid circular import at top of class
        coordinates = self.sources.cartesian
        r = self.sources.vertical_polar[:, 2].mean()
        target = self._get_coordinates((azimuth, elevation, r), 'spherical').cartesian
        # compute distances from target direction
        distances = numpy.sqrt(((target - coordinates)**2).sum(axis=1))
        if method == 'nearest':
            idx_nearest = numpy.argmin(distances)
            filt = self[idx_nearest]
        else:
            # triangulate source positions into triangles
            if not scipy:
                raise ImportError('Need scipy.spatial for barycentric interpolation.')
            tri = scipy.spatial.ConvexHull(coordinates)
            if plot_tri:
                ax = plt.subplot(projection='3d')
                for simplex in tri.points[tri.simplices]:
                    polygon = Poly3DCollection([simplex])
                    polygon.set_color(numpy.random.rand(3))
                    ax.add_collection3d(polygon)
                    mins = coordinates.min(axis=0)
                    maxs = coordinates.max(axis=0)
                    xlim, ylim, zlim = list(zip(mins, maxs))
                    ax.set_xlim(xlim)
                    ax.set_ylim(ylim)
                    ax.set_zlim(zlim)
                    ax.set_xlabel('X [m]')
                    ax.set_ylabel('Y [m]')
                    ax.set_zlabel('Z [m]')
                    plt.show()
            # for each simplex, find the coordinates, test if target in triangle (by finding minimal d)
            d_min = numpy.inf
            for i, vertex_list in enumerate(tri.simplices):
                simplex = tri.points[vertex_list]
                d, a = HRTF._barycentric_weights(simplex, target)
                if d < d_min:
                    d_min, idx, weights = d, i, a
            vertex_list = tri.simplices[idx]
            # we now have the indices of the filters and the corresponding weights
            amplitudes = list()
            for idx in vertex_list:
                freqs, amps = self[idx].tf(show=False)  # get their transfer functions
                amplitudes.append(amps)  # we could interpolate here if frequencies differ between filters
            avg_amps = amplitudes[0] * weights[0] + amplitudes[1] * weights[1] + amplitudes[2] * weights[2]  # average
            gains = avg_amps - avg_amps.max()  # shift so that maximum is zero, because we can only attenuate
            gains[gains < -60] = -60  # limit dynamic range to 60 dB
            gains_lin = 10**(gains/20)  # transform attenuations in dB to factors
            filt_l = Filter.band(frequency=list(freqs), gain=list(gains_lin[:, 0]), length=self[idx].n_samples, fir='IR',
                                 samplerate=self[vertex_list[0]].samplerate)
            filt_r = Filter.band(frequency=list(freqs), gain=list(gains_lin[:, 1]), length=self[idx].n_samples, fir='IR',
                                 samplerate=self[vertex_list[0]].samplerate)
            filt = Filter(data=[filt_l, filt_r])
            itds = list()
            for idx in vertex_list:
                taps = Binaural(self[idx])  # recast filter taps as Binaural sound
                itds.append(taps.itd())  # use Binaural.itd to compute correlation lag between channels
            avg_itd = itds[0] * weights[0] + itds[1] * weights[1] + itds[2] * weights[2]  # average ITD
            filt = filt.delay(avg_itd / self.samplerate)
        return filt

    @staticmethod
    def _barycentric_weights(triangle, point):
        '''
        Returns:
            (None | numpy.array): barycentric weights for a given triangle (array of coordinates of points) and target
            point IF the point is inside the triangle; None if the point is outside the triangle.
        '''
        # compute barycentric weights via the area of the 3 triangles formed between target and 3 nearest sources:
        dist = lambda p1, p2: numpy.sqrt(((p1 - p2)**2).sum())
        d1 = dist(point, triangle[0, :])  # distances from target to each source
        d2 = dist(point, triangle[1, :])
        d3 = dist(point, triangle[2, :])
        d12 = dist(triangle[0, :], triangle[1, :])  # distance between sources 1 and 2
        d13 = dist(triangle[0, :], triangle[2, :])
        d23 = dist(triangle[1, :], triangle[2, :])
        # compute triangle areas from length of sides (distances) with Heron's Formula
        a = numpy.array([0., 0., 0.])
        p = (d2 + d3 + d23) / 2
        a[0] = numpy.sqrt(p * (p-d2) * (p-d3) * (p-d23))
        p = (d1 + d3 + d13) / 2
        a[1] = numpy.sqrt(p * (p-d1) * (p-d3) * (p-d13))
        p = (d1 + d2 + d12) / 2
        a[2] = numpy.sqrt(p * (p-d1) * (p-d2) * (p-d12))
        p = (d12 + d13 + d23) / 2
        tot = numpy.sqrt(p * (p-d12) * (p-d13) * (p-d23))
        return a.sum() - tot, a / a.sum()  # normalize by total area = barycentric weights of sources in idx_triangle

    def vsi(self, sources=None, equalize=True):
        """
        Compute  the "vertical spectral information" which is a measure of the dissimilarity of spectral profiles
        at different elevations. The vsi relates to behavioral localization accuracy in the vertical dimension
        (Trapeau and Schönwiesner, 2016). It is computed as one minus the average of the correlation coefficients
        between all combinations of directional transfer functions of the specified `sources`. A set of identical
        transfer functions results in a vsi of 0 whereas highly different transfer functions will result in a high VSI
        (empirical maximum is ~1.07, KEMAR has a VSI of 0.82).

        Arguments:
            sources (None | list): indices of sources for which to compute the VSI. If None use the vertical midline.
            equalize (bool): If True, apply the `diffuse_field_equalization` method (set to False if the hrtf object
                is already diffuse-field equalized).
        Returns:
            (float): the vertical spectral information between the specified `sources`.
        """
        if sources is None:
            sources = self.cone_sources()
        if equalize:
            dtf = self.diffuse_field_equalization()
            tfs = dtf.tfs_from_sources(sources=sources)
        else:
            tfs = self.tfs_from_sources(sources=sources)
        sum_corr = 0
        n = 0
        for i in range(len(sources)):
            for j in range(i+1, len(sources)):
                sum_corr += numpy.corrcoef(tfs[:, i], tfs[:, j])[1, 0]
                n += 1
        return 1 - sum_corr / n

    def plot_sources(self, idx=None, show=True, label=False, axis=None):
        """
        Plot source locations in 3D.

        Arguments:
            idx (list of int): indices to highlight in the plot
            show (bool): whether to show plot (set to False if plotting into an axis and you want to add other elements)
            label (bool): if True, show the index of each source in self.sources as text label, if idx is also given,
                then only theses sources are labeled
            axis (mpl_toolkits.mplot3d.axes3d.Axes3D): axis to draw the plot on
        """
        if matplotlib is False or Axes3D is False:
            raise ImportError('Plotting 3D sources requires matplotlib and mpl_toolkits')
        if axis is None:
            ax = plt.subplot(projection='3d')
        else:
            if not isinstance(axis, Axes3D):
                raise ValueError("Axis must be instance of Axes3D!")
            ax = axis
        coordinates = self.sources.cartesian
        ax.scatter(coordinates[:, 0], coordinates[:, 1], coordinates[:, 2], c='b', marker='.')
        ax.axes.set_xlim3d(left=numpy.min(coordinates), right=numpy.max(coordinates))
        if label and idx is None:
            for i in range(coordinates.shape[0]):
                ax.text(coordinates[i, 0], coordinates[i, 1], coordinates[i, 2], f'{i}', size=8, zorder=1, color='k')
        ax.scatter(0, 0, 0, c='r', marker='o')
        if self.listener:
            x_, y_, z_, u, v, w = zip(*[self.listener['viewvec'], self.listener['upvec']])
            ax.quiver(x_, y_, z_, u, v, w, length=0.5, colors=['r', 'b', 'r', 'r', 'b', 'b'])
        if idx is not None:
            ax.scatter(coordinates[idx, 0], coordinates[idx, 1], coordinates[idx, 2], c='r', marker='o')
            if label:
                for i in idx:
                    ax.text(coordinates[i, 0], coordinates[i, 1], coordinates[i, 2], f'{i}', size=8, zorder=1, color='k')
        ax.set_xlabel('X [m]')
        ax.set_ylabel('Y [m]')
        ax.set_zlabel('Z [m]')
        if show:
            plt.show()

    @staticmethod
    def kemar():
        '''
        Provides HRTF data from the KEMAR recording (normal pinna) conducted by Gardner and Martin at MIT in 1994
        (MIT Media Lab Perceptual Computing - Technical Report #280) and converted to the SOFA Format. Slab includes a
        compressed copy of the data. This function reads it and returns the corresponding HRTF object. The objects is
        cached in the class variable `_kemar` and repeated calls return the cached object instead of reading the file
        from disk again.

        Returns:
            (slab.HRTF): the KEMAR HRTF data.
        '''
        global _kemar
        if _kemar is None:
            kemar_path = pathlib.Path(__file__).parent.resolve() / pathlib.Path('data') / 'mit_kemar_normal_pinna.bz2'
            _kemar = pickle.load(bz2.BZ2File(kemar_path, "r"))
            _kemar.sources = HRTF._get_coordinates(_kemar.sources, 'spherical')
            _kemar.datatype = 'FIR'
            for f in _kemar.data:
                f.fir = 'IR'
        return _kemar

    @staticmethod
    def estimate_hrtf(recordings, signal, sources, listener=None):
        """
        Compute a set of transfer functions from binaural recordings and an input (reference) signal.
        For each sound source, compute the DFT of left- and right-ear recordings
        and divide by the Fourier transform of the input signal to obtain the head related transfer function.

        Arguments:
            signal (slab.Signal | slab.Sound): the signal used to produce the in-ear recordings.
            recordings (list): in-ear recordings stored in a list of slab.Binaural objects.
            sources (numpy.array): interaural polar coordinates (azimuth, elevation, distance) of all sources,
                number and order of sources must match the recordings.

        Returns:
            (slab.HRTF): an HRTF object with the dimensions specified by the recordings and the source file.
        """
        if not isinstance(recordings, list) or not isinstance(recordings[0], Sound):
            raise ValueError('Recordings must be provided as a list of slab.Sound objects.')
        if isinstance(sources, (list, tuple)):
            sources = numpy.array(sources)
        if len(sources.shape) == 1:  # a single location (vector) needs to be converted to a 2d matrix
            sources = sources[numpy.newaxis, ...]
        if len(sources) != len(recordings):
            raise ValueError('Number of sound sources must be equal to number of recordings.')
        rec_samplerate = recordings[0].samplerate
        rec_n_samples = recordings[0].n_samples
        rec_data = []
        for recording in recordings:
            if not (recording.n_channels == 2 and recording.n_samples == recordings[0].n_samples
                    and recording.samplerate == rec_samplerate):
                raise ValueError('Number of channels, samples and samplerate must be equal for all recordings.')
            rec = copy.deepcopy(recording)
            # rec.data -= numpy.mean(rec.data, axis=0)  # remove DC component in FFT of recordings
            rec_data.append(rec.data.T)
        rec_data = numpy.asarray(rec_data)
        if not signal.samplerate == rec_samplerate:
            signal = signal.resample(rec_samplerate)
        if not signal.n_samples == rec_n_samples:
            sig_freq_bins = numpy.fft.rfftfreq(signal.n_samples, d=1 / signal.samplerate)
            rec_freq_bins = numpy.fft.rfftfreq(rec_n_samples, d=1 / rec_samplerate)
            sig_fft = numpy.interp(rec_freq_bins, sig_freq_bins, numpy.fft.rfft(signal.data[:, 0]))
        else:
            sig_fft = numpy.fft.rfft(signal.data[:, 0])
        with numpy.errstate(divide='ignore'):
            hrtf_data = numpy.fft.rfft(rec_data) / sig_fft
        if not listener:
            listener = {'pos': numpy.array([0., 0., 0.]), 'view': numpy.array([1., 0., 0.]),
                        'up': numpy.array([0., 0., 1.]), 'viewvec': numpy.array([0., 0., 0., 1., 0., 0.]),
                        'upvec': numpy.array([0., 0., 0., 0., 0., 1.])}
        return HRTF(data=numpy.abs(hrtf_data), datatype='TF', samplerate=rec_samplerate, sources=sources,
                    listener=listener)

    def write_sofa(self, filename):
        """
        Save the HRTF data to a SOFA file.

        Arguments:
            filename (str | pathlib.Path): path, the file is written to.
        """
        if isinstance(filename, pathlib.Path):
            filename = str(filename)
        if pathlib.Path(filename).is_file():
            pathlib.Path(filename).unlink()  # overwrite if filename already exists
        if netCDF4 is False:
            raise ImportError('Writing sofa files requires netCDF4.')
        sofa = netCDF4.Dataset(filename, 'w', format='NETCDF4')  # Create SOFA file
        # ----------Dimensions----------#
        m = self.n_sources  # number of measurements (= n_sources)
        n = self[0].n_samples  # n_samples - frequencies of fourier filter or taps of FIR filter
        r = 2  # number of receivers (HRTFs measured for 2 ears)
        e = 1  # number of emitters (1 speaker per measurement)
        i = 1  # always 1
        c = 3  # number of dimensions in space (elevation, azimuth, radius)
        sofa.createDimension('M', m)
        sofa.createDimension('N', n)
        sofa.createDimension('E', e)
        sofa.createDimension('R', r)
        sofa.createDimension('I', i)
        sofa.createDimension('C', c)
        # ----------Attributes----------#
        sofa.DataType = self.datatype
        if self.datatype == 'FIR':
            sofa.SOFAConventions, sofa.SOFAConventionsVersion = 'SimpleFreeFieldHRIR', '2.0'
            delayVar = sofa.createVariable('Data.Delay', 'f8', ('I', 'R'))
            delay = numpy.zeros((i, r))
            delayVar[:, :] = delay
            dataIRVar = sofa.createVariable('Data.IR', 'f8', ('M', 'R', 'N'))
            dataIRVar.ChannelOrdering = 'acn'
            dataIRVar.Normalization = 'sn3d'
            IR_data = []
            for idx in numpy.asarray(self[:]):
                IR_data.append(idx.T)
            dataIRVar[:] = numpy.asarray(IR_data)
        elif self.datatype == 'TF':
            sofa.SOFAConventions, sofa.SOFAConventionsVersion = 'SimpleFreeFieldHRTF', '2.0'
            dataRealVar = sofa.createVariable('Data.Real', 'f8', ('M', 'R', 'N'))  # data
            TF_data = []
            for idx in numpy.asarray(self[:]):
                TF_data.append(idx.T)
            dataRealVar[:] = numpy.asarray(TF_data)
            dataImagVar = sofa.createVariable('Data.Imag', 'f8', ('M', 'R', 'N'))
            dataImagVar[:] = numpy.zeros((m, r, n))  # for internal use, store real data only
            NVar = sofa.createVariable('N', 'f8', ('N'))
            NVar.LongName = 'frequency'
            NVar.Units = 'hertz'
            NVar[:] = n
        sofa.RoomType = 'free field'
        sofa.Conventions, sofa.Version = 'SOFA', '2.0'
        sofa.APIName, sofa.APIVersion = 'pysofaconventions', '0.1'
        sofa.AuthorContact, sofa.License = 'Leipzig University', 'PublicLicence'
        sofa.ListenerShortName, sofa.Organization = 'sub01', 'Eurecat - UPF'
        sofa.DateCreated, sofa.DateModified = str(datetime.datetime.now()), str(datetime.datetime.now())
        sofa.Title, sofa.DatabaseName = 'sofa_title', 'UniLeipzig Freefield'
        # ----------Variables----------#
        listenerPositionVar = sofa.createVariable('ListenerPosition', 'f8', ('I', 'C'))
        listenerPositionVar.Units = 'metre'
        listenerPositionVar.Type = 'cartesian'
        listenerPositionVar[:] = self.listener['pos']
        receiverPositionVar = sofa.createVariable('ReceiverPosition', 'f8', ('R', 'C', 'I'))
        receiverPositionVar.Units = 'metre'
        receiverPositionVar.Type = 'cartesian'
        receiverPositionVar[:] = numpy.zeros((r, c, i))
        sourcePositionVar = sofa.createVariable('SourcePosition', 'f8', ('M', 'C'))
        sourcePositionVar.Units = 'degree, degree, metre'
        sourcePositionVar.Type = 'spherical'
        sourcePositionVar[:] = self.sources.vertical_polar  # array of speaker positions
        emitterPositionVar = sofa.createVariable('EmitterPosition', 'f8', ('E', 'C', 'I'))
        emitterPositionVar.Units = 'metre'
        emitterPositionVar.Type = 'cartesian'
        emitterPositionVar[:] = numpy.zeros((e, c, i))
        listenerUpVar = sofa.createVariable('ListenerUp', 'f8', ('I', 'C'))
        listenerUpVar.Units = 'metre'
        listenerUpVar.Type = 'cartesian'
        listenerUpVar[:] = self.listener['up']
        listenerViewVar = sofa.createVariable('ListenerView', 'f8', ('I', 'C'))
        listenerViewVar.Units = 'metre'
        listenerViewVar.Type = 'cartesian'
        listenerViewVar[:] = self.listener['view']
        samplingRateVar = sofa.createVariable('Data.SamplingRate', 'f8', ('I'))
        samplingRateVar.Units = 'hertz'
        samplingRateVar[:] = self.samplerate
        sofa.close()


class Room:
    """
    Class for acoustic room simulations, including binaural room impulse responses for given source and listener positions,
    echo locations, reverberation, and wall filters. Initializing a room objects immediately computes echo image locations.
    Reverb times, a reverb filter, and the binaural room impulse response can then be obtained with the respective methods.

    Arguments:
        size (list | numpy.ndarray): width, length, and height of room in meters.
        listener (list | numpy.ndarray): cartesian coordinates of listener in the room (x,y,z)
        source (list | numpy.ndarray): source location relative to listener in spherical coordiates (azimuth,elevation,distance).
            The source location is immediately transformed into cartesian coordinates relative to the room.
        order (int): Number of reflections to simulate (default 2).
        max_echos (int): Number of source images to keep (defaults to simulated number, around 50 is perceptually sufficient).
        absorption (int | list): absorption coefficients [wall, [floor, [ceiling]]] in 1/m^2
        wall_filters (int | list): [wall, [floor, [ceiling]]] filter indices into the database !!NOT IMPLEMENTED YET!!

    Attributes:
        .image_locs (numpy.ndarray): Echo locations in spherical coordinates (azimuth, elevation, distance).
        .orders (list): Reflection order for each echo in image_locs.
        .floor_orders (list): Number of reflections from the floor for each echo.
        .ceil_orders (list): Number of reflections from the ceiling for each echo.

    Example:
        import slab
    """

    def __init__(self, size=[4,5,3], listener=[2,2.5,1.5], source=[0,0,1.4], order=2, max_echos=50, absorption=[0.1,], wall_filters=None):
        self.size = size
        self.listener = numpy.array(listener)
        if numpy.any(self.listener < 0) or numpy.any(self.listener > size):
            raise ValueError('Listener outside room bounds!')
        self.order = order
        self.absorption = absorption
        self.max_echos = max_echos
        self.wall_filters = wall_filters
        self.set_source(source)

    def __repr__(self):
        return f'{type(self)} (\n{repr(self.size)}\n{repr(self.listener)}\n{repr(self.source)}\n{repr(self.order)}\n{repr(self.absorption)}'

    def __str__(self):
        return f'{type(self)} with dimensions {self.size}, listener at {self.listener} and source at {self.source[0]}'

    def set_source(self, source):
        """
        Set the source attribute, convert source to cartesian coordinates, and recompute image_locs and orders.

        Arguments:
            source (list | numpy.ndarray): source location relative to listener in spherical coordiates (azimuth,elevation,distance).
        """
        source_loc_cartesian = HRTF._get_coordinates(source, 'spherical')[0]
        source_loc_cartesian += self.listener
        if numpy.any(source_loc_cartesian < 0) or numpy.any(source_loc_cartesian > self.size):
            raise ValueError('Source outside room bounds!')
        self.source = source_loc_cartesian
        self.image_locs, self.orders, self.floor_orders, self.ceil_orders = \
                    Room._simulated_echo_locations(self.size, self.listener, source_loc_cartesian, self.order, self.max_echos)

    @staticmethod
    def _simulated_echo_locations(room_size=[4,5,3], listener_loc=[2,2.5,1.5], source_loc=[2,4,1.5], sim_ord=2, max_sources=None):
        """
        Compute the locations and order of echos in a shoebox room using the mirror image method.

        Arguments:
            room_size (list | numpy.ndarray): width, length, and height of room in meters.
            listener_loc (list | numpy.ndarray): cartesian coordinates of listener in the room (x,y,z)
            source_loc (list | numpy.ndarray): cartesian coordinates of sound source in the room (x,y,z)
            sim_ord (int): Number of reflections to simulate (default 2).
            max_sources (int): Number of source images to keep (defaults to simulated number, around 50 is perceptually sufficient).
        Returns:
            img_loc_pol (numpy.ndarray): Echo locations in spherical coordinates (azimuth, elevation, distance).
            order (list): Reflection order for each echo in img_pol_loc.
            floor_order (list): Number of reflections from the floor for each echo.
            ceil_order (list): Number of reflections from the ceiling for each echo.
        """
        # calculate source images
        Lx, Ly, Lz = room_size
        Sx, Sy, Sz = source_loc[0,0], source_loc[0,1], source_loc[0,2]
        Ix = []
        Iy = []
        Iz = []
        order = []
        floor_order = []
        ceil_order = []
        # positions of images, all combinations of mirrored locs along x,y,z axes
        for (l, m, n) in itertools.product(range(-sim_ord,sim_ord+1), repeat=3):
            xl =  Sx + 2 * l * Lx
            xr = -Sx + 2 * l * Lx
            Ix.extend([xl,xl,xl,xl,xr,xr,xr,xr])
            yf =  Sy + 2 * m * Ly
            yb = -Sy + 2 * m * Ly
            Iy.extend([yf,yf,yb,yb,yf,yf,yb,yb])
            zu =  Sz + 2 * n * Lz
            zd = -Sz + 2 * n * Lz
            Iz.extend([zu,zd,zu,zd,zu,zd,zu,zd])
        for x, y, z in zip(Ix, Iy, Iz):
            if x >= 0:
                n_order = numpy.floor(x/Lx)
            else:
                n_order = numpy.ceil(-x/Lx)
            if y >= 0:
                n_order = n_order + numpy.floor(y/Ly)
            else:
                n_order = n_order + numpy.ceil(-y/Ly)
            if z >= 0:
                n_order = n_order + numpy.floor(z/Lz)
                n_floor_order = numpy.floor(numpy.floor(z/Lz)/2)
                n_ceil_order = numpy.ceil(numpy.floor(z/Lz)/2)
            else:
                n_order = n_order + numpy.ceil(-z/Lz)
                n_floor_order = numpy.ceil(numpy.ceil(-z/Lz)/2)
                n_ceil_order = numpy.floor(numpy.ceil(-z/Lz)/2)
            order.append(int(n_order))
            floor_order.append(int(n_floor_order))
            ceil_order.append(int(n_ceil_order))
        img_locs = numpy.stack([Ix, Iy, Iz], axis=-1)
        # source images are in cartesian, with reference to room coordinate
        # transform it into vertical polar with reference to listener
        _, img_loc_pol, _ = HRTF._get_coordinates(img_locs - listener_loc, 'cartesian')
        if max_sources is None:
            max_sources = len(order)
        # sort according to distance; get the index used to sort
        s_idx = numpy.argsort(img_loc_pol[:, 2])
        # choose N closest sources
        img_loc_pol = img_loc_pol[s_idx[:max_sources]]
        order = numpy.array(order)[s_idx[:max_sources]]
        floor_order = numpy.array(floor_order)[s_idx[:max_sources]]
        ceil_order = numpy.array(ceil_order)[s_idx[:max_sources]]
        return img_loc_pol, order, floor_order, ceil_order

    @staticmethod
    def _walltrns(dry_filter, total_order, floor_order):
        """
        Applies wall and floor absorptions to the HRTF filter for one echo direction.

        Arguments:
            dry_filter (slab.Filter): head-related transfer function for a given direction in the form of a binaural IR Filter.
            total_order (int): total number of reflections (echo order)
            floor_order (int): number of floor reflections
        Returns:
            (slab.Filter): 2-channel filter (left/right ear)
        """
        floorcoef = numpy.array([0.6921,    0.0523,    0.0612,    0.0020,    0.0071,    0.0071,    0.0162,
                              0.0176,    0.0187,    0.0152,    0.0117,    0.0075,    0.0045,    0.0025,
                              0.0014,    0.0008,    0.0004,    0.0002,    0.0001,   -0.0000,   -0.0001,
                              -0.0002,   -0.0002,   -0.0003,   -0.0003,   -0.0003,  -0.0003,   -0.0002,
                              -0.0002,   -0.0001,   -0.0001,   -0.0001,   -0.0001,  -0.0001,   -0.0001,
                              -0.0001,   -0.0001,   -0.0001,   -0.0000,   -0.0000,   0.0000,    0.0000,
                              0.0000,    0.0000,    0.0001,    0.0001,    0.0001,    0.0001,    0.0001,
                              0.0001,    0.0001,    0.0001,    0.0001,    0.0001,    0.0001,    0.0001,
                              0.0001,    0.0001,    0.0001,    0.0001,    0.0001,    0.0001,    0.0001,
                              0.0001,    0.0001,    0.0001,    0.0001,    0.0001]) # Uncut pile Carpet, fabric impregnated a6-1
        wallcoef = numpy.array([0.2655,    0.3718,    0.0672,   -0.0008,    0.0259,    0.0207,    0.0195,
                              0.0141,    0.0120,    0.0090,    0.0082,    0.0068,    0.0064,    0.0057,
                              0.0051,    0.0046,    0.0041,    0.0037,    0.0033,    0.0030,    0.0027,
                              0.0024,    0.0022,    0.0019,    0.0017,    0.0016,    0.0014,    0.0013,
                              0.0012,    0.0010,    0.0009,    0.0009,    0.0008,    0.0007,    0.0006,
                              0.0006,    0.0005,    0.0005,    0.0004,    0.0004,    0.0004,    0.0003,
                              0.0003,    0.0003,    0.0002,    0.0002,    0.0002,    0.0002,    0.0001,
                              0.0001,    0.0001]) # Coconut Fibre - Roller felt a3.1-1
        HRIR_l = dry_filter[:, 0]
        HRIR_r = dry_filter[:, 1]
        for n in range(total_order - floor_order):
            HRIR_l = numpy.convolve(HRIR_l, wallcoef)
            HRIR_r = numpy.convolve(HRIR_r, wallcoef)
        for n in range(floor_order):
            HRIR_l = numpy.convolve(HRIR_l, floorcoef)
            HRIR_r = numpy.convolve(HRIR_r, floorcoef)
        return Filter([HRIR_l, HRIR_r], samplerate=dry_filter.samplerate, fir='IR')

    def reverb_time(self, size=None, absorption=None):
        """
        Use Sabine formula to calculate reverberation time for the given room.
        Can be called as static method arguments size and absorption, otherwise uses current room parameters.

        Arguments:
            size (list): optional, room dimensions in meters (width, length, height)
            absorption (list): optional, absorption coefficients [wall, [floor, [ceiling]]] in 1/m^2
        Returns:
            (float): reverberation time in s
        """
        c = 344 # speed of sound
        if size is None:
            size = self.size
        if absorption is None:
            absorption = self.absorption
        vol = size[0] * size[1] * size[2]
        wall_surface = (size[0] * size[2]) * 2 + (size[1] * size[2]) * 2
        floor_surface = size[0] * size[1]
        ceil_surface = floor_surface
        # calculate sabins
        if len(self.absorption) == 3:
            sa = wall_surface * absorption[0] + floor_surface * absorption[1] + \
                 ceil_surface * absorption[2]
        elif len(self.absorption) == 2:
            sa = wall_surface * absorption[0] + 2 * floor_surface * absorption[1]
        elif len(self.absorption) == 1:
            sa = (wall_surface + 2 * floor_surface) * absorption[0]
        else:
            raise ValueError("absorption coefficients not understood")
        # sabine formula
        return 24 * numpy.log(10) / c * vol / sa

    def reverb(self, t_reverb=None, low_factor=0.8, trim=0.25, samplerate=44100):
        """
        Generate an exponentially decaying reverberation tail impulse response.

        Arguments:
            t_reverb (int | float): optional, reverberation time in seconds (default: use current room parameters)
            low_factor (float): t_reverb * low_factor is the reverb time used for the low frequencies (default 0.8)
            trim (int | float): shorten the (usually unnecessarily long) reverb tail; if int: trim to that number of samples, if float < 1:  trim to that fraction (default is 0.5).
            samplerate (int | float): samplerate of the resulting noise (default 44100)
        Returns:
            (slab.Filter): reverberation tail impulse response
        """
        if t_reverb is None:
            t_reverb = self.reverb_time()
        RTlow, RThigh = low_factor * t_reverb, t_reverb
        RT = max(RTlow, RThigh)
        t = numpy.arange(RT * samplerate * 2) / samplerate
        envLow = numpy.exp(-t/RTlow * 60/20 * numpy.log(10))
        envHigh = numpy.exp(-t/RThigh * 60/20 * numpy.log(10))
        n_samples = len(envHigh)
        # generate filtered noise
        noise_data = numpy.random.randn(n_samples)
        # fft params
        fft_length = 2 ** (len(noise_data) - 1).bit_length()
        f = samplerate * (numpy.arange(fft_length / 2) + 1) / fft_length
        fr = f[::-1]
        ff = numpy.hstack([f, fr])
        # log-powered filter in frequency domain
        s = (numpy.log2(ff) - numpy.log2(ff[0])) / (numpy.log2(ff[int(len(ff)/2)]) - numpy.log2(ff[0]))
        out = numpy.zeros((len(envHigh), 2))
        # convolution
        nd_fft = numpy.fft.fft(noise_data, fft_length)
        out[:, 0] = numpy.fft.ifft(nd_fft * s).real[:len(noise_data)] * envHigh + \
                    numpy.fft.ifft(nd_fft * (1 - s)).real[:len(noise_data)] * envLow
        noise_data = numpy.random.randn(n_samples)
        nd_fft = numpy.fft.fft(noise_data, fft_length)
        out[:, 1] = numpy.fft.ifft(nd_fft * s).real[:len(noise_data)] * envHigh + \
                    numpy.fft.ifft(nd_fft * (1 - s)).real[:len(noise_data)] * envLow
        out = Filter(out, fir='IR', samplerate=samplerate)
        if isinstance(trim, float):
            trim = int(out.n_samples * trim)
        out = out.resize(trim)
        return out

    def hrir(self, reverb=None, hrtf=None, trim=0.5):
        """
        Compute the binaural room response. The resulting filter has the same samplerate as the HRTF object.

        Arguments:
            reverb (float | slab.Filter): if None, generate the decaying reverberation tail from current room parameters (default),
                                        if float, generate the reverberation tail with a time constant of 'reverb',
                                        if slab.Filter, use the provided binaural filter
            hrtf (slab.HRTF): HRTF dataset (default: slab.HRTF.kemar)
            trim (int | float): shorten the (usually unnecessarily long) reverb tail; if int: trim to that number of samples, if float < 1:  trim to that fraction (default is 0.5).

        Return:
            (slab.Filter): binaural room response for given source, room, and listener position
        """
        if hrtf is None:
            hrtf = HRTF.kemar() # get HRIRs from slab
        if reverb is None:
            reverb = self.reverb(samplerate=hrtf.samplerate)
        elif isinstance(reverb, float): # use as reverb time
            reverb = self.reverb(t_reverb=reverb, samplerate=hrtf.samplerate)
        delays = self.image_locs[:, 2] / 344 # source distance * speed of sound
        # delays[1:] = delays[1:] + ((numpy.random.rand(len(delays)-1)-.5) * delays[1:] / 10)
        onsets = numpy.ceil(delays * hrtf.samplerate).astype(int)
        onsets[onsets==0] = 1
        r0 = min(delays) # delay of the direct sound
        n_samples = numpy.max(onsets) + hrtf[0].n_taps + 1000
        echo_HRIR = numpy.zeros((n_samples, 2))
        for i, (azi, ele) in enumerate(self.image_locs[:,:2]):
            this_dry_hrtf = hrtf.interpolate(azimuth=azi, elevation=ele) # get the hrtf for the current echo direction
            single_echo_HRIR = Room._walltrns(this_dry_hrtf, self.orders[i], self.floor_orders[i])
            single_echo_HRIR *= (r0 / delays[i]) # attenuate sound pressure by distance (p ~ 1/d)
            # update total HRIR
            offset = onsets[i] + single_echo_HRIR.n_taps
            echo_HRIR[onsets[i]:offset, :] += single_echo_HRIR
        # combine calculated sources and simulated tail
        peak = numpy.max(numpy.abs(single_echo_HRIR)) # maximum amplitude of last (faintest) echo
        combined_HRIR = numpy.zeros((onsets[i] + reverb.n_taps, 2))
        combined_HRIR[onsets[i]:, :] = reverb * peak # put reverbtail at start of last echo
        combined_HRIR[:offset] += echo_HRIR[:offset]
        out = Filter(combined_HRIR, samplerate=hrtf.samplerate, fir='IR')
        if isinstance(trim, float):
            trim = int(out.n_samples * trim)
        out = out.resize(trim)
        return out



import sys
import os
import pathlib

__version__ = '1.6.1'

# The variable _in_notebook is used to enable audio playing in Jupiter notebooks
# and on Google colab (see slab.sound.play())
try:
    shell = get_ipython().__class__.__module__
    if 'terminal' in shell:
        _in_notebook = False
    else:
        _in_notebook = True  # probably in a notebook
except NameError:
    _in_notebook = False  # probably standard Python interpreter

sys.path.append('..\\')

from slab.hrtf import HRTF, Room
from slab.psychoacoustics import *
from slab.binaural import Binaural
from slab.sound import Sound, set_default_level, set_calibration_intensity, get_calibration_intensity, calibrate
from slab.signal import Signal, set_default_samplerate, get_default_samplerate
from slab.filter import Filter

def cite(fmt='bibtex'):
    """
    Return the citation string of the slab module.

    Arguments:
        fmt (str): if bibtex, return the bibtex citation string, otherwise return the text reference
    """
    if fmt == 'bibtex':
        return """@article{Schönwiesner2021, doi = {10.21105/joss.03284}, url = {https://doi.org/10.21105/joss.03284}, year = {2021}, publisher = {The Open Journal}, volume = {6}, number = {62}, pages = {3284}, author = {Marc Schönwiesner and Ole Bialas}, title = {s(ound)lab: An easy to learn Python package for designing and running psychoacoustic experiments.}, journal = {Journal of Open Source Software}}"""
    return """Schönwiesner et al., (2021). s(ound)lab: An easy to learn Python package for designing and running psychoacoustic experiments. Journal of Open Source Software, 6(62): 3284, https://doi.org/10.21105/joss.03284"""



import copy
import warnings
import numpy
try:
    import matplotlib
    import matplotlib.pyplot as plt
except ImportError:
    matplotlib, plt = False, False
try:
    import scipy.signal
except ImportError:
    scipy = False

_default_samplerate = 8000  # default samplerate in Hz; used by all methods if on samplerate argument is provided.

def set_default_samplerate(samplerate):
    global _default_samplerate
    _default_samplerate = samplerate

def get_default_samplerate():
    global _default_samplerate
    return _default_samplerate

class Signal:
    """
    Base class for Signal data (from which the Sound and Filter class inherit).
    Provides arithmetic operations, slicing, and conversion between samples and times.

    Arguments:
        data (numpy.ndarray | slab.Signal | list): samples of the sound. If it is an array, the first dimension
            should represent the number of samples and the second one the number of channels. If it's an object,
            it must have a .data attribute containing an array. If it's a list, the elements can be arrays or objects.
            The output will be a multi-channel sound with each channel corresponding to an element of the list.
        samplerate (int | None): the samplerate of the sound. If None, use the default samplerate.
    Attributes:
        .duration: duration of the sound in seconds
        .n_samples: duration of the sound in samples
        .n_channels: number of channels in the sound
        .times: list with the time point of each sample
    Examples::

        import slab, numpy
        sig = slab.Signal(numpy.ones([10,2]),samplerate=10)  # create a sound
        sig[:5] = 0  # set the first 5 samples to 0
        sig[:,1]  # select the data from the second channel
        sig2 = sig * 2  # multiply each sample by 2
        sig_inv = -sig  # invert the phase
    """
    # instance properties
    n_samples = property(fget=lambda self: self.data.shape[0],
                         doc='The number of samples in the Signal.')
    duration = property(fget=lambda self: self.data.shape[0] / self.samplerate,
                        doc='The length of the Signal in seconds.')
    times = property(fget=lambda self: numpy.arange(self.data.shape[0], dtype=float) / self.samplerate,
                     doc='An array of times (in seconds) corresponding to each sample.')
    n_channels = property(fget=lambda self: self.data.shape[1],
                          doc='The number of channels in the Signal.')

    # __methods (class creation, printing, and slice functionality)
    def __init__(self, data, samplerate=None):
        if hasattr(data, 'samplerate') and samplerate is not None:
            warnings.warn('First argument has a samplerate property. Ignoring given samplerate.')
        if isinstance(data, numpy.ndarray):
            self.data = numpy.array(data, dtype='float')
        elif isinstance(data, (list, tuple)):
            if all([hasattr(c, 'data') and hasattr(c, 'samplerate') for c in data]):  # all slab objects
                if not all(c.samplerate == data[0].samplerate for c in data):
                    raise ValueError('All elements of list must have the same samplerate!')
                if not all(c.n_samples == data[0].n_samples for c in data):
                    raise ValueError('All elements of list must have the same number of samples!')
                channel_data = [c.data for c in data]  # all clear, now get channel arrays
                self.data = numpy.hstack(channel_data)
                self.samplerate = data[0].samplerate
            elif all([isinstance(c, numpy.ndarray) for c in data]):
                channel_data = tuple(Signal(c, samplerate=samplerate) for c in data)
                self.data = numpy.hstack(channel_data)
        # any object with data and samplerate attributes can be recast as Signal
        elif hasattr(data, 'data') and hasattr(data, 'samplerate'):
            self.data = data.data
            self.samplerate = data.samplerate
        else:
            raise TypeError('Cannot initialise Signal with data of class ' + str(data.__class__))
        if self.data.ndim == 1:
            self.data = self.data[:,numpy.newaxis]
        elif self.data.shape[1] > self.data.shape[0]:
            if not len(data) == 0:  # dont transpose if data is an empty array
                self.data = self.data.T
        if not hasattr(self, 'samplerate'):  # if samplerate has not been set, use default
            if samplerate is None:
                self.samplerate = _default_samplerate
            else:
                self.samplerate = samplerate

    def __repr__(self):
        return f'{type(self)} (\n{repr(self.data)}\n{repr(self.samplerate)})'

    def __str__(self):
        return f'{type(self)} duration {self.duration}, samples {self.n_samples}, channels {self.n_channels},' \
               f'samplerate {self.samplerate}'

    def _repr_html_(self):
        'HTML image representation for Jupyter notebook support'
        elipses = '\u2026'
        class_name = str(type(self))[8:-2]
        html = [f'<h4>{class_name} with samplerate = {self.samplerate}</h4>']
        html += ['<table><tr><th>#</th>']
        samps, chans = self.data.shape
        html += (f'<th>channel {j}</th>' for j in range(chans))
        if samps > 7:
            rows = [0, 1, 2, -1, samps-3, samps-2, samps-1]  # -1 will output '...'
        else:
            rows = range(samps)
        for i in rows:
            html.append(f'<tr><th>{elipses if i == -1 else i}</th>')
            for j in range(chans):
                if i == -1:
                    html.append(f'<td>{elipses}</td>')
                else:
                    html.append(f'<td>{self.data[i, j]:.5f}</td>')
        html.append('</table>')
        return ''.join(html)

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __setitem__(self, key, value):
        return self.data.__setitem__(key, value)

    # arithmetic operators
    def __add__(self, other):
        new = copy.deepcopy(self)
        if isinstance(other, type(self)):
            new.data = self.data + other.data
        else:
            new.data = self.data + other
        return new
    __radd__ = __add__

    def __sub__(self, other):
        new = copy.deepcopy(self)
        if isinstance(other, type(self)):
            new.data = self.data - other.data
        else:
            new.data = self.data - other
        return new
    __rsub__ = __sub__

    def __mul__(self, other):
        new = copy.deepcopy(self)
        if isinstance(other, type(self)):
            new.data = self.data * other.data
        else:
            new.data = self.data * other
        return new
    __rmul__ = __mul__

    def __truediv__(self, other):
        new = copy.deepcopy(self)
        if isinstance(other, type(self)):
            new.data = self.data / other.data
        else:
            new.data = self.data / other
        return new
    __rtruediv__ = __truediv__

    def __pow__(self, other):
        new = copy.deepcopy(self)
        new.data = self.data ** other
        return new

    def __neg__(self):
        new = copy.deepcopy(self)
        new.data = self.data*-1
        return new

    def __len__(self):
        return self.n_samples

    # static methods (belong to the class, but can be called without creating an instance)
    @staticmethod
    def in_samples(ctime, samplerate):
        """
        Converts time values in seconds to samples. This is used to enable input in either samples (integers) or
        seconds (floating point numbers) in the class.

        Arguments:
            ctime (int | float | list | numpy.ndarray): the time(s) to convert to samples.
            samplerate (int): the samplerate of the sound.
        Returns:
             (int | list | numpy.ndarray): the time(s) in samples.
        """
        out = None
        if isinstance(ctime, (int, numpy.int64)):  # single int is treated as samples
            out = ctime
        elif isinstance(ctime, (float, numpy.float64)):
            out = int(numpy.rint(ctime * samplerate))
        elif isinstance(ctime, numpy.ndarray):
            if isinstance(ctime[0], numpy.int64):
                out = ctime
            elif isinstance(ctime[0], numpy.float64):
                out = numpy.int64(numpy.rint(ctime * samplerate))
            else:
                ValueError('Unsupported type in numpy.ndarray (must be int64 of float 64)')
        elif isinstance(ctime, (list, tuple)):  # convert one by one
            out = numpy.empty_like(ctime)
            for i, c in enumerate(ctime):
                out[i] = numpy.int64(Signal.in_samples(c, samplerate))
        else:
            ValueError(
                'Unsupported type for ctime (must be int, float, numpy.ndarray of ints or floats, list or tuple)!')
        return out

    @staticmethod
    def set_default_samplerate(samplerate):
        """ Sets the global default samplerate for Signal objects, by default 8000 Hz. """
        global _default_samplerate
        _default_samplerate = samplerate
        return _default_samplerate

    # instance methods (belong to instances created from the class)
    def channel(self, n):
        """
        Get a single data channel.

        Arguments:
            n (int): channel index
        Returns:
            (slab.Signal): a new instance of the class that contains the selected channel as data.
        """
        new = copy.deepcopy(self)
        new.data = self.data[:, n]
        new.data.shape = (len(new.data), 1)
        return new

    def channels(self):
        """ Returns generator that yields channel data as objects of the calling class. """
        return (self.channel(i) for i in range(self.n_channels))

    def resize(self, duration):
        """
        Change the duration by padding with zeros or cutting the data.

        Arguments:
            duration (float | int): new duration of the sound in seconds (given a float) or in samples (given an int).
        Returns:
            (slab.Signal): a new instance of the same class with the specified duration.
        """
        duration = Signal.in_samples(duration, self.samplerate)
        resized = copy.deepcopy(self)
        if duration == len(self.data):
            pass  # already correct size
        elif duration < len(self.data):  # cut the data
            resized.data = self.data[:duration, :]
        else:  # pad with zeros
            padding = numpy.zeros((duration - len(self.data), self.n_channels))
            resized.data = numpy.concatenate((self.data, padding))
        return resized

    def trim(self, start=0, stop=None):
        """ Trim the signal by returning the section between `start` and `stop`.
        Arguments:
            start (float | int): start of the section in seconds (given a float) or in samples (given an int).
            stop (float | int): end of the section in seconds (given a float) or in samples (given an int).
        Returns:
            (slab.Signal): a new instance of the same class with the specified duration. """
        start = Signal.in_samples(start, self.samplerate)
        if stop is None:
            stop = self.n_samples - 1
        else:
            stop = Signal.in_samples(stop, self.samplerate)
        if stop >= self.n_samples:
            stop = self.n_samples - 1
        if start < 0:
            start = 0
        if start >= stop:
            raise ValueError('Start must precede stop.')
        trimmed = copy.deepcopy(self)
        trimmed.data = self.data[start:stop, :]
        return trimmed

    def resample(self, samplerate):
        """
        Resample the sound.

        Arguments:
            samplerate (int): the samplerate of the resampled sound.
        Returns:
            (slab.Signal): a new instance of the same class with the specified samplerate.
        """
        if scipy is False:
            raise ImportError('Resampling requires scipy.sound.')
        if self.samplerate == samplerate:
            return self
        out = copy.deepcopy(self)
        new_n_samples = int(numpy.rint(samplerate*self.duration))
        new_signal = numpy.zeros((new_n_samples, self.n_channels))
        for chan in range(self.n_channels):
            new_signal[:, chan] = scipy.signal.resample(
                self.channel(chan), new_n_samples).flatten()
        out.data = new_signal
        out.samplerate = samplerate
        return out

    def envelope(self, apply_envelope=None, times=None, kind='gain', cutoff=50):
        """
        Either apply an envelope to a sound or, if no `apply_envelope` was specified, compute the Hilbert envelope
        of the sound.

        Arguments:
            apply_envelope (None | numpy.ndarray): data to multiply with the sound. The envelope is linearly
                interpolated to be the same length as the sound. If None, compute the sound's Hilbert envelope
            times (None | numpy.ndarray | list): If None a vector linearly spaced from 0 to the duration of the sound
                is used. If time points (in seconds, clamped to the the sound duration) for the amplitude values
                in envelope are supplied, then the interpolation is piecewise linear between pairs of time and envelope
                valued (must have same length).
            kind (str): determines the unit of the envelope value
            cutoff (int): Frequency of the lowpass filter that is applied to remove the temporal fine structure in Hz.
        Returns:
            (slab.Signal): Either a copy of the instance with the specified envelope applied or the sound's
                Hilbert envelope.
        """
        if apply_envelope is None:  # get the signals envelope
            return self._get_envelope(kind, cutoff)
        else:  # apply the envelope to the sound
            return self._apply_envelope(apply_envelope, times, kind)

    def _get_envelope(self, kind, cutoff):
        if scipy is False:
            raise ImportError('Calculating envelopes requires scipy.signal.')
        envs = numpy.abs(scipy.signal.hilbert(self.data, axis=0))
        # 50Hz lowpass filter to remove fine-structure
        filt = scipy.signal.firwin(1000, cutoff, pass_zero=True, fs=self.samplerate)
        envs = scipy.signal.filtfilt(filt, [1], envs, axis=0)
        envs[envs <= 0] = numpy.finfo(float).eps  # remove negative values and zeroes
        if kind == 'dB':
            envs = 20 * numpy.log10(envs)  # convert amplitude to dB
        elif not kind == 'gain':
            raise ValueError('Kind must be either "gain" or "dB"!')
        return Signal(envs, samplerate=self.samplerate)

    def _apply_envelope(self, envelope, times, kind):
        # TODO: write tests for the new options!
        if hasattr(envelope, 'data') and hasattr(envelope, 'samplerate'): # it's a child object
            envelope = envelope.resample(samplerate=self.samplerate)
        elif isinstance(envelope, (list, numpy.ndarray)):  # make it a Signal, so we can use .n_samples and .n_channels
            envelope = Signal(numpy.array(envelope), samplerate=self.samplerate)
        new = copy.deepcopy(self)
        if times is None:
            times = numpy.linspace(0, 1, envelope.n_samples) * self.duration
        else:  # times vector was supplied
            if len(times) != envelope.n_samples:
                raise ValueError('Envelope and times need to be of equal length!')
            times = numpy.array(times)
            times[times > self.duration] = self.duration  # clamp between 0 and sound duration
            times[times < 0] = 0
        envelope_interp = numpy.empty((self.n_samples,envelope.n_channels)) # prep array for interpolated env
        for i in range(envelope.n_channels): # interp each channel
            envelope_interp[:,i] = numpy.interp(self.times, times, envelope[:,i])
        if kind == 'dB':
            envelope_interp = 10**(envelope_interp/20.) # convert dB to gain factors
        if envelope.n_channels == new.n_channels: # corresponding chans -> just multiply
            new.data *= envelope_interp
        elif envelope.n_channels == 1 and new.n_channels > 1: # env 1 chan, sound multichan -> apply env to each chan
            for i in range(new.n_channels):
                new.data[:,i] *= envelope_interp[:,0]
        else: # or neither (raise error)
            raise ValueError('Envelope needs to be 1d or have the same number of channels as the sound!')
        return new

    def delay(self, duration=1, channel=0, filter_length=2048):
        """
        Add a delay to one channel.

        Arguments:
            duration (int | float | array-like): duration of the delay in seconds (given a float) or samples (given
                an int). Given an array with the same length as the sound, each sample is delayed by the
                corresponding number of seconds. This option is used by in `slab.Binaural.itd_ramp`.
            channel (int): The index of the channel to add the delay to
            filter_length (int): Must be and even number. determines the accuracy of the reconstruction when
                using fractional sample delays. Defaults to 2048, or the sound length for shorter signals.
        Returns:
            (slab.Signal): a copy of the instance with the specified delay.
        """
        new = copy.deepcopy(self)
        if channel >= self.n_channels:
            raise ValueError('Channel must be smaller than number of channels in sound!')
        if filter_length % 2:
            raise ValueError('Filter_length must be even!')
        if self.n_samples < filter_length:  # reduce the filter_length to the sound length of short signals
            filter_length = self.n_samples - 1 if self.n_samples % 2 else self.n_samples  # make even
        center_tap = int(filter_length / 2)
        t = numpy.array(range(filter_length))
        if isinstance(duration, (int, float, numpy.int64, numpy.float64)):  # just a constant delay
            duration = Signal.in_samples(duration, self.samplerate)
            if duration > self.n_samples:
                raise ValueError("Duration of the delay cant be greater longer then the sound!")
            x = t - duration + 1
            window = 0.54 - 0.46 * numpy.cos(2 * numpy.pi * (x+0.5) /
                                             filter_length)  # Hamming window
            if numpy.abs(duration) < 1e-10:
                tap_weight = numpy.zeros_like(t)
                tap_weight[center_tap] = 1
            else:
                tap_weight = window * numpy.sinc(x-center_tap)
            new.data[:, channel] = numpy.convolve(self.data[:, channel], tap_weight, mode='same')
        else:  # dynamic delay
            if len(duration) != self.n_samples:
                ValueError('Duration shorter or longer than sound!')
            duration *= self.samplerate  # assuming vector in seconds, convert to samples
            padding = numpy.zeros(center_tap)
            # for zero-padded convolution (potential edge artifacts!)
            sig = numpy.concatenate((padding, new.channel(channel), padding), axis=None)
            for i, current_delay in enumerate(duration):
                x = t-current_delay
                window = 0.54 - 0.46 * numpy.cos(2 * numpy.pi *
                                                 (x+0.5) / filter_length)  # Hamming window
                if numpy.abs(current_delay) < 1e-10:
                    tap_weight = numpy.zeros_like(t)
                    tap_weight[center_tap] = 1
                else:
                    tap_weight = window * numpy.sinc(x-center_tap)
                    sig_portion = sig[i:i+filter_length]
                    # sig_portion and tap_weight have the same length, so the valid part of the convolution is just
                    # one sample, which gets written into the sound at the current index
                    new.data[i, channel] = numpy.convolve(sig_portion, tap_weight, mode='valid')[0]
        return new

    def plot_samples(self, show=True, axis=None):
        """
        Stem plot of the samples of the signal.

        Arguments:
            show (bool): whether to show the plot right after drawing.
            axis (matplotlib.axes.Axes | None): axis to plot to. If None create a new plot.
        """
        if matplotlib is False:
            raise ImportError('Plotting signals requires matplotlib.')
        if axis is None:
            _, axis = plt.subplots()
        if self.n_channels == 1:
            axis.stem(self.channel(0))
        else:
            for i in range(self.n_channels):
                axis.stem(self.channel(i), label=f'channel {i}', linefmt=f'C{i}')
            plt.legend()
        axis.set(title='Samples', xlabel='Number', ylabel='Value')
        if show:
            plt.show()



from setuptools import setup, find_packages
import re

with open('README.md') as f:
    readme = f.read()

# extract version
with open('slab/__init__.py') as file:
    for line in file.readlines():
        m = re.match("__version__ *= *['\"](.*)['\"]", line)
        if m:
            version = m.group(1)

setup(name='slab',
      version=version,
      description='Tools for generating and manipulating digital signals, particularly sounds.',
      long_description=readme,
      long_description_content_type='text/markdown',
      url='http://github.com/DrMarc/slab.git',
      author='Marc Schoenwiesner',
      author_email='marc.schoenwiesner@gmail.com',
      license='MIT',
      python_requires='>=3.6',
      install_requires=['numpy', 'scipy', 'SoundFile', 'SoundCard',
        "matplotlib < 3.4; python_version == '3.6'",
        "matplotlib; python_version >= '3.7'"],
      extras_require={'testing': ['pytest', 'h5netcdf'],
                      'docs': ['sphinx', 'sphinx-rtd-theme'],
                      'hrtf': ['h5netcdf']},
      packages=find_packages(),
      package_data={'slab': ['data/mit_kemar_normal_pinna.bz2']},
      include_package_data=True,
      zip_safe=False)



import slab
from itertools import zip_longest
import tempfile
import numpy
from matplotlib import pyplot as plt
import random
from pathlib import Path
DIR = tempfile.TemporaryDirectory()
PATH = Path(DIR.name)
plt.ioff()
# NOTE: Everything involving pressing a key is currently untested because curses does not run within pytest


def test_sequence_from_trials():
    for i in range(100):  # generate from integers
        start = numpy.random.randint(1, 100)
        stop = start+numpy.random.randint(1, 10)
        trials = [numpy.random.randint(start, stop) for i in range(100)]
        sequence = slab.Trialsequence(trials=trials)
        assert all(numpy.unique(sequence.trials) == numpy.array(range(1, sequence.n_conditions+1)))
    trials = ["a", "x", "x", "z", "a", "a", "a", "z"]
    sequence = slab.Trialsequence(trials=trials)
    assert all(numpy.unique(sequence.trials) == numpy.array(range(1, sequence.n_conditions + 1)))
    sounds = [slab.Sound.pinknoise(), slab.Sound.whitenoise()]
    trials = [random.choice(sounds) for i in range(50)]
    sequence = slab.Trialsequence(trials=trials)
    assert all(numpy.unique(sequence.trials) == numpy.array(range(1, sequence.n_conditions + 1)))


def test_sequence():
    for _ in range(100):
        conditions_list = [numpy.random.randint(2, 10), ["a", "b", "c"], [("a", "b"), (1.5, 3.2)],
                           [slab.Sound.pinknoise(), slab.Sound.whitenoise()]]
        kinds = ["random_permutation", "non_repeating", "infinite"]
        for conditions in conditions_list:
            n_reps = numpy.random.randint(1, 10)
            kind = random.choice(kinds)
            sequence = slab.Trialsequence(conditions=conditions, n_reps=n_reps, kind=kind)
            if isinstance(conditions, int):
                conditions = list(range(1, conditions+1))
            assert sequence.conditions == conditions
            assert all(numpy.unique(sequence.trials) == numpy.array(range(1, sequence.n_conditions + 1)))
            if kind != "infinite":
                assert sequence.n_trials == len(conditions) * n_reps
                for trial in sequence:
                    assert trial == sequence.conditions[sequence.trials[sequence.this_n]-1]
            else:
                count = 0
                for trial in sequence:
                    assert trial == sequence.conditions[sequence.trials[sequence.this_n]-1]
                    count += 1
                    if count > 100:
                        break


def test_deviants():
    for i in range(100):
        conditions = 4
        n_reps = 50
        n_trials = conditions*n_reps
        deviant_frequency = 0.2 * random.random() + .05
        sequence = slab.Trialsequence(conditions=4, n_reps=50, deviant_freq=deviant_frequency)
        count_deviants = 0
        for trial in sequence:
            if trial == 0:
                count_deviants += 1
            else:
                assert trial == sequence.conditions[sequence.trials[sequence.this_n] - 1]
        assert count_deviants == sequence.trials.count(0) == int(n_trials*deviant_frequency)


def test_staircase():  # this seems to block
    stairs1 = slab.Staircase(start_val=10, n_reversals=4)
    stairs2 = slab.Staircase(start_val=8, n_reversals=6)
    stairs = zip_longest(stairs1, stairs2)
    count = 0
    for stim1, stim2 in stairs:
        count += 1
        if count > 100:
            break
        print(count)
        if stim1 is not None:
            r1 = stairs1.simulate_response(4)
            stairs1.add_response(r1)
            # stairs1.print_trial_info()
        if stim2 is not None:
            r2 = stairs2.simulate_response(2)
            stairs2.add_response(r2)
            # stairs2.print_trial_info()
    fig, ax = plt.subplots(2)
    stairs1.plot(axis=ax[0], show=False)
    stairs2.plot(axis=ax[1], show=False)
    # adaptive staircase
    stairs = slab.Staircase(start_val=10, n_reversals=18, step_sizes=[4, 1])
    for stimulus_value in stairs:
        response = stairs.simulate_response(threshold=3)
        stairs.add_response(response)
    stairs.save_csv(PATH / "staircase.csv")


def test_precomputed():
    sounds = [slab.Sound.whitenoise() for _ in range(10)]
    sounds = slab.Precomputed(sounds)
    sounds = slab.Precomputed(sounds.random_choice(5))
    sounds.write(PATH / "precomputed.zip")
    sounds = slab.Precomputed.read(PATH / "precomputed.zip")


def test_results():
    slab.psychoacoustics.results_folder = PATH
    results = slab.ResultsFile(subject="MrPink")
    for data in [[1, 2, 3], slab.Trialsequence()]:
        results.write(data)
    results.read()
    results = slab.ResultsFile.read_file(slab.ResultsFile.previous_file(subject="MrPink"))
    results.clear()
    # ResultsTable
    results = slab.ResultsTable(subject="MrPink", columns="subject, trial")
    row = results.Row(subject=results.subject, trial=1)
    results.write(row)



""" psychoacoustics exports classes for handling psychophysical procedures and
measures, like trial sequences and staircases."""

import io
import pathlib
import datetime
import json
import pickle
import zipfile
import collections
from contextlib import contextmanager
from abc import abstractmethod
import warnings
import matplotlib.cbook # necessary for matplotlib versions <3.5 to suppress a MatplotlibDeprecationWarning
try:
    import curses
except ImportError:
    curses = None
import numpy
try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None
import slab

results_folder = 'Results'
input_method = 'keyboard'  #: sets the input for the Key context manager to 'keyboard', 'buttonbox', 'prompt', or 'figure'


class _Buttonbox:
    """
    Adapter class to allow easy switching between input from the keyboard via curses (no ENTER keypress needed after
    single button press) to the Python input prompt, which can be used with an external button box (custom arduino device
    that sends keystrokes) or when running from a Jupiter or Colab notebook.
    """
    @staticmethod
    def getch():
        input_key = input()  # buttonbox has to return the key followed by ENTER!
        if input_key:
            return ord(input_key)

class _FigChar:
    """
    Adapter class to allow easy switching to input via the current_character attribute of stairs figure.
    Set slab.psychoacoustics.input_method = 'figure' to use. A figure with the name 'stairs' will be opened if it is not
    already present. If used together with the plot method of the Staircase class, input is acquired through the stairs
    plot. Depending on the operating system, you may have to click once into the figure to give it focus.
    """
    warnings.filterwarnings("ignore", category=matplotlib.MatplotlibDeprecationWarning)

    @staticmethod
    def getch():
        global fig_key
        def _on_key(event):
            global fig_key
            fig_key = event.key
        fig = plt.figure('stairs')
        _ = fig.canvas.mpl_connect('key_press_event', _on_key)
        fig_key = None # reset
        while not fig_key:
            plt.pause(0.01) # wait for 10ms, but keep figure event loop running
        return ord(fig_key)


@contextmanager
def key(mesg=None):
    """
    Wrapper for curses module to simplify getting a single keypress from the terminal (default), a buttonbox, or a
    figure. Set slab.psychoacoustics.input_method = 'buttonbox' to use a custom USB buttonbox, to 'figure' to open
    a figure called 'stairs' (if not already opened by the `slab.Staricase.plot` method), or to 'prompt' for a
    simple Python prompt (requires pressing Enter/Return after the response). Optionally takes a string argument
    which is printed in the terminal for conveying instructions to the participant.

    Example::

        with slab.key('Waiting for buttons 1 (yes) or 2 (no).') as key:
        response = key.getch()
    """

    if input_method == 'keyboard':
        if curses is None:
            raise ImportError(
                'You need curses to use the keypress class (pip install curses (or windows-curses))')
        curses.filter()
        stdscr = curses.initscr()
        curses.noecho()
        curses.cbreak()
        stdscr.clear()
        stdscr.refresh()
        if mesg is not None:
            stdscr.addstr(str(mesg))
        yield stdscr
        curses.nocbreak()
        curses.echo()
        curses.endwin()
    elif input_method == 'buttonbox' or input_method == 'prompt':
        if mesg is not None:
            print(mesg)
        yield _Buttonbox
    elif input_method == 'figure':
        if mesg is not None:
            print(mesg)
        yield _FigChar
    else:
        raise ValueError('Unknown input method!')


class LoadSaveMixin:
    """ Mixin to provide loading and saving functions. Supports JSON the pickle format """

    def save_pickle(self, file_name, clobber=False):
        """
        Save the object as pickle file.

        Arguments:
            file_name (str | pathlib.Path): name of the file to create.
            clobber (bool): overwrite existing file with the same name, defaults to False.
        Returns:
            (bool): True if writing was successful.
        """
        if isinstance(file_name, pathlib.Path):
            file_name = str(file_name)
        if pathlib.Path(file_name).exists() and not clobber:
            raise FileExistsError("Select clobber=True to overwrite.")
        with open(file_name, 'wb') as fp:
            pickle.dump(self.__dict__, fp, protocol=pickle.HIGHEST_PROTOCOL)
            return True

    def load_pickle(self, file_name):
        """
        Read pickle file and deserialize the object into `self.__dict__`.

        Attributes:
            file_name (str | pathlib.Path): name of the file to read.
        """
        if isinstance(file_name, pathlib.Path):
            file_name = str(file_name)
        with open(file_name, 'rb') as fp:
            self.__dict__ = pickle.load(fp)

    def save_json(self, file_name=None, clobber=False):
        """
        Save the object as JSON file. The object's __dict__ is serialized and saved in standard JSON format, so that it
        can be easily reconstituted (see load_json method). Raises FileExistsError if the file exists, unless `clobber`
        is True. When `file_name` in None (default), the method returns the JSON string, in case you want to inspect it.
        Note that Numpy arrays are not serializable and are converted to Python int. This works because the
        Trialsequence and Staircase classes use arrays of indices. If your instances of these classes contain arrays of
        float, use `save_pickle` instead.

        Arguments:
            file_name (str | pathlib.Path): name of the file to create. If None or 'stdout', return a JSON object.
            clobber (bool): overwrite existing file with the same name, defaults to False.
        Returns:
            (bool): True if writing was successful.
        """
        def default(i): return int(i) if isinstance(i, (numpy.int64,numpy.int32)) else i  # helper for converting numpy arrays
        if isinstance(file_name, pathlib.Path):
            file_name = str(file_name)
        if (file_name is None) or (file_name == 'stdout'):
            return json.dumps(self.__dict__, indent=2, default=default)
        if pathlib.Path(file_name).exists() and not clobber:
            raise FileExistsError("Select clobber=True to overwrite.")
        try:
            with open(file_name, 'w') as f:
                json.dump(self.__dict__, f, indent=2, default=default)
                return True
        except (TypeError, ValueError):  # type error caused by json dump, value error by default function
            print("Your sequence contains data which is not JSON serializable, use the save_pickle method instead.")

    def load_json(self, file_name):
        """
        Read JSON file and deserialize the object into `self.__dict__`.

        Attributes:
            file_name (str | pathlib.Path): name of the file to read.
        """
        if isinstance(file_name, pathlib.Path):
            file_name = str(file_name)
        with open(file_name, 'r') as f:
            self.__dict__ = json.load(f)


class TrialPresentationOptionsMixin:
    """
    Mixin to provide alternative forced-choice (AFC) and Same-Different trial presentation methods and
    response simulation to `Trialsequence` and `Staircase`.
    """
    @abstractmethod
    def add_response(self, response):
        pass

    @abstractmethod
    def print_trial_info(self):
        pass

    def present_afc_trial(self, target, distractors, key_codes=(range(49, 58)), isi=0.25, print_info=True):
        """
        Present the reference and distractor sounds in random order and acquire a response keypress.
        The subject has to identify at which position the reference was played. The result (True if response was correct
        or False if response was wrong) is stored in the sequence via the `add_response` method.

        Arguments:
            target (instance of slab.Sound): sound that ought to be identified in the trial
            distractors (instance or list of slab.Sound): distractor sound(s)
            key_codes (list of int): ascii codes for the response keys (get code for button '1': ord('1') --> 49)
                pressing the second button in the list is equivalent to the response "the reference was the second sound
                played in this trial". Defaults to the key codes for buttons '1' to '9'
            isi (int or float): inter stimulus interval which is the pause between the end of one sound and the start
            of the next one.
            print_info (bool): If true, call the `print_trial_info` method afterwards
        """
        if isinstance(distractors, list):
            stims = [target] + distractors  # assuming sound object and list of sounds
        else:
            stims = [target, distractors]  # assuming two sound objects
        order = numpy.random.permutation(len(stims))
        for idx in order:
            stim = stims[idx]
            stim.play()
            plt.pause(isi)
        with key() as k:
            response = k.getch()
        interval = numpy.where(order == 0)[0][0]
        interval_key = key_codes[interval]
        response = response == interval_key
        self.add_response(response)
        if print_info:
            self.print_trial_info()

    def present_tone_trial(self, stimulus, correct_key_idx=0, key_codes=(range(49, 58)), print_info=True):
        """
        Present the reference and distractor sounds in random order and acquire a response keypress.
        The result (True if response was correct or False if response was wrong) is stored in the sequence via the
        `add_response` method.

        Arguments:
            stimulus (slab.Sound): sound played in the trial.
            correct_key_idx (int | list of int): index of the key in `key_codes` that represents a correct response.
                Response is correct if `response == key_codes[correct_key_idx]`. Can be a list of ints if several keys
                are counted as correct response.
            key_codes (list of int): ascii codes for the response keys (get code for button '1': ord('1') --> 49).
            print_info (bool): If true, call the `print_trial_info` method afterwards.
        """
        stimulus.play()
        with slab.key() as k:
            response = k.getch()
        response = response == key_codes[correct_key_idx]
        self.add_response(response)
        if print_info:
            self.print_trial_info()

    def simulate_response(self, threshold=None, transition_width=2, intervals=1, hitrates=None):
        """
        Return a simulated response to the current condition index value by calculating the hitrate from a
        psychometric (logistic) function. This is only sensible if trials is numeric and an interval scale representing
        a continuous stimulus value.

        Arguments:
            threshold(None | int | float): Midpoint of the psychometric function for adaptive testing. When the
                intensity of the current trial is equal to the `threshold` the hitrate is 50 percent.
            transition_width (int | float): range of stimulus intensities over which the hitrate increases
                from 0.25 to 0.75.
            intervals (int): use 1 (default) to indicate a yes/no trial, 2 or more to indicate an alternative forced
             choice trial. The number of choices determines the probability for a correct response by chance.
            hitrates (None | list | numpy.ndarray): list or numpy array of hitrates for the different conditions,
                to allow custom rates instead of simulation. If given, `threshold` and `transition_width` are not used.
                If a single value is given, this value is used.
        """
        slope = 0.5 / transition_width
        if isinstance(self, slab.psychoacoustics.Trialsequence): # check which class the mixin is in
            current_condition = self.trials[self.this_n]
        elif isinstance(self, slab.psychoacoustics.Staircase):
            current_condition = self._next_intensity
        else:
            return None
        if hitrates is None:
            if threshold is None:
                raise ValueError("threshold can't be None if hitrates is None!")
            hitrate = 1 / (1 + numpy.exp(4 * slope * (threshold - current_condition)))  # scale/4  = slope at midpoint
        else:
            if isinstance(hitrates, (list, numpy.ndarray)):
                hitrate = hitrates[current_condition]
            else:
                hitrate = hitrates
        hit = numpy.random.rand() < hitrate  # True with probability hitrate
        if hit or intervals == 1:
            return hit
        return numpy.random.rand() < 1/intervals  # still 1/intervals chance to hit the right interval


class Trialsequence(collections.abc.Iterator, LoadSaveMixin, TrialPresentationOptionsMixin):
    """
    Randomized, non-adaptive trial sequences.

    Arguments:
        conditions (list | int | str): defines the different stimuli appearing the sequence. If given a list,
            every element is one condition. The elements can be anything - strings, dictionaries, objects etc.
            Note that, if the elements are not JSON serializable, the sequence can only be saved as a pickle file.
            If conditions is an integer i, the list of conditions is given by range(i). A string is treated as the
            filename of a previously saved trial sequence object, which is then loaded.
        n_reps (int): number of repetitions for each condition. Number of trials is given by len(conditions)*n_reps).
        trials (None | list | numpy.ndarray): The sequence of trials in the order in which they are appearing in
            sequence. Defaults to None, because trials are usually generated by the class based on the other
            parameters. However, it is possible to pass a list or one-dimensional array. In that case the parameters
            for generating the sequence are ignored.
        kind (str): The kind of randomization used to generate the trial sequence. Possible options are:
            `non_repeating` (randomization without direct repetition of a condition, default if n_conditions > 2),
            `random_permutation` (complete randomization, default if `n_conditions` <= 2) or
            `infinite` (sequence that reset when reaching the end to generate an infinite number of trials.
            randomization method is random_permutation` if n_conditions` <= 2 and `non_repeating` otherwise).
        deviant_freq (float): frequency with which deviants (encoded as 0) appear in the sequence. The minimum number
            of trials between two deviants is 3 if deviant frequency is below 10%, 2 if it is below 20% and 1 if it
            is below 30%. A deviant frequency greater than 30% is not supported
        label (str): a text label for the sequence.
    Attributes:
        .trials: the order in which the conditions are repeated in the sequence. The elements are integers referring
             to indices in `conditions`, starting from 1. 0 represents a deviant (only present if `deviant_freq` > 0)
        .n_trials: the total number of trials in the sequence
        .conditions: list of the different unique elements in the sequence
        .n_conditions: number of conditions, is equal to len(conditions) or len(conditions)+1 if there are deviants
        .n_remaining: the number of trials remaining i.e. that have not been called when iterating trough the sequence
        .this_n: current trials index in the entire sequence, equals the number of trials completed so far
        .this_trial: a dictionary giving the parameters of the current trial
        .finished: boolean signaling if all trials have been called
        .kind: randomization kind of sequence (`random_permutation`, `non_repeating`, `infinite`)
        .data: list with the same length as the one in the `trials` attribute. On sequence generation, `data` is a
            list of empty lists. Then , one can use the `add_response` method to append to the list belonging to the
            current trial
    """
    def __init__(self, conditions=2, n_reps=1, trials=None, kind=None, deviant_freq=None, label=''):
        self.label = label
        self.n_reps = int(n_reps)
        if isinstance(conditions, pathlib.Path):
            conditions = str(conditions)
        if isinstance(conditions, str):
            if not pathlib.Path(conditions).exists():
                raise ValueError(f"could not load the file {conditions}")
            try:
                self.load_json(conditions)  # import entire object from file
            except (UnicodeDecodeError, json.JSONDecodeError) as _:
                self.load_pickle(conditions)
        else:
            if isinstance(conditions, int):
                self.conditions = list(range(1, conditions+1))
            else:
                self.conditions = conditions
            self.n_conditions = len(self.conditions)
            if trials is None:  # generate stimulus sequence
                if kind is None:
                    kind = 'random_permutation' if self.n_conditions <= 2 else 'non_repeating'
                if kind == 'random_permutation':
                    self.trials = self._create_random_permutation(self.n_conditions, self.n_reps)
                elif kind == 'non_repeating':
                    self.trials = self._create_simple_sequence(self.n_conditions, self.n_reps)
                elif kind == 'infinite':
                    # implementation if infinite sequence is a bit of a hack (number of completed trials needs
                    # to be calculated as: trials.this_rep_n * trials.n_conditions + trials.this_trial_n + 1)
                    # It's also not possible to make an infinite sequence with deviants.
                    if deviant_freq is not None:
                        raise ValueError("Deviants are not implemented for infinite sequences!")
                    if self.n_conditions <= 2:
                        self.trials = self._create_random_permutation(self.n_conditions, 5)
                        self.n_reps = 5
                    else:
                        self.trials = self._create_simple_sequence(self.n_conditions, 1)
                        self.n_reps = 1
                else:
                    raise ValueError(f'Unknown kind parameter: {kind}!')
                if deviant_freq is not None:  # insert deviants
                    deviants = slab.Trialsequence._deviant_indices(n_standard=int(self.n_conditions * n_reps),
                                                                   deviant_freq=deviant_freq)
                    self.trials = numpy.insert(arr=self.trials, obj=deviants, values=0)
                    self.n_conditions += 1  # add one condition for deviants
            else:  # make a sequence from a given list of trials
                self.conditions = list(set(trials))
                for i, condition in enumerate(
                        self.conditions):  # encode conditions as integers 1 to n_conditions in trials
                    for t, trial in enumerate(trials):
                        if trial == condition:
                            trials[t] = i + 1
                self.trials = trials
                self.n_conditions = len(self.conditions)
            if isinstance(self.trials, numpy.ndarray):
                self.trials = self.trials.tolist()  # convert trials to list
            self.this_n = -1  # trial index in entire sequence
            self.this_trial = []  # condition of current trial
            self.finished = False
            self.data = []  # holds responses if TrialPresentationOptions methods are called
            self.n_trials = len(self.trials)
            self.n_remaining = self.n_trials
            self.kind = kind
            self.data = [[] for _ in self.trials]

    def __repr__(self):
        return self.__dict__.__repr__()

    def __str__(self):
        return f'Trialsequence, trials {"inf" if self.kind=="infinite" else self.n_trials}, ' \
               f'remaining {"inf" if self.kind=="infinite" else self.n_remaining}, current condition {self.this_trial}'

    def __next__(self):
        """
        Is called when iterating trough a sequenceAdvances to next trial and returns it. Updates attributes
        `this_trial` and `this_n`. If the trials have ended this method will raise a StopIteration.
        Returns:
             (int): current element of the list in `trials`
        """
        self.this_n += 1
        self.n_remaining -= 1
        if self.n_remaining < 0:  # all trials complete
            if self.kind == 'infinite':  # finite sequence -> reset and start again
                # new sequence, avoid start with previous condition
                self.trials = self._create_simple_sequence(len(self.conditions), self.n_reps,
                                                           dont_start_with=self.trials[-1])
                self.this_n = 0
                self.n_remaining = self.n_trials - 1  # reset trial countdown to length of new trial
                #  sequence (subtract 1 because we return the 0th trial below)
            else:  # finite sequence -> finish
                self.this_trial = []
                self.finished = True
        if self.finished:
            raise StopIteration
        if self.trials[self.this_n] == 0:
            self.this_trial = 0
        else:
            self.this_trial = self.conditions[self.trials[self.this_n]-1]  # fetch the trial info
        return self.this_trial

    def add_response(self, *response):
        """
        Append response(s) to the list in the `data` attribute belonging to the current trial (see Trialsequence doc).

        Attributes:
             response (any): data to append to the list. Can be anything but save_json method won't be available if
                the content of `response` is not JSON serializable (if it's an object for example).
        """
        if self.this_n < 0:
            print("Can't add response because trial hasn't started yet!")
        else:
            for r in response:
                self.data[self.this_n].append(r)

    def print_trial_info(self):
        """ Convenience method for printing current trial information. """
        print(f'{self.label} | trial # {self.this_n} of {"inf" if self.kind=="infinite" else self.n_trials} '
              f'({"inf" if self.kind=="infinite" else self.n_remaining} remaining): condition {self.this_trial}, '
              f'last response: {self.data[self.this_n-1]}')

    @staticmethod
    def _create_simple_sequence(n_conditions, n_reps, dont_start_with=None):
        """
        Create a randomized sequence of integers without direct repetitions of any element.

        Arguments:
            n_conditions (int): the number of conditions in the list. The array returned contains integers from 1
                to the value of `n_conditions`.
            n_reps (int): number that each element is repeated. Length of the returned array is `n_conditions * n_reps`
            dont_start_with (int): if not None, dont start the sequence with this integer. Can be useful if several
                sequences are used and the final trial of the last sequence should not be the same as the first
                element of the next sequence.
        Returns:
            (numpy.ndarray): randomized sequence of length n_conditions * n_reps without direct repetitions of any
            element.
        """
        permute = list(range(1, n_conditions+1))
        if dont_start_with is not None:
            trials = [dont_start_with]
        else:
            trials = []
        for _ in range(n_reps):
            numpy.random.shuffle(permute)
            if len(trials) > 0:
                while trials[-1] == permute[0]:
                    numpy.random.shuffle(permute)
            trials += permute
        if dont_start_with is not None: # delete first entry ('dont_start_with')
            trials = trials[1:]
        return numpy.array(trials)

    @staticmethod
    def _deviant_indices(n_standard, deviant_freq=.1):
        """
        Create sequence for an oddball experiment which contains two conditions: standards (1) and deviants (0).

        Arguments:
            n_standard (int): number of standard trials, encoded as 1, in the sequence.
            deviant_freq (float): frequency of deviants, encoded as 0, in the sequence. Also determines the minimum
            number of standards between two deviants which is 3 if deviant_freq <= .1, 2 if deviant_freq <= .2 and
            1 if deviant_freq <= .3. A deviant frequency > .3 is not supported.
        Returns:
            (numpy.ndarray): sequence of length n_standard+(n_standard*deviant_freq) with deviants.
        """
        if deviant_freq <= .1:
            min_dist = 3
        elif deviant_freq <= .2:
            min_dist = 2
        elif deviant_freq <= .3:
            min_dist = 1
        else:
            raise ValueError("Deviant frequency can't be greater than 0.3!")
        # get the possible combinations of deviants and normal trials:
        n_deviants = int(n_standard*deviant_freq)
        indices = range(n_standard)
        deviant_indices = numpy.random.choice(indices, n_deviants, replace=False)
        deviant_indices.sort()
        dist = numpy.diff(deviant_indices)
        while numpy.min(dist) < min_dist:  # reshuffle until minimum distance is satisfied
            deviant_indices = numpy.random.choice(indices, n_deviants, replace=False)
            deviant_indices.sort()
            dist = numpy.diff(deviant_indices)
        return deviant_indices

    @staticmethod
    def _create_random_permutation(n_conditions, n_reps):
        """
        Create a completely random sequence of integers.

        Arguments:
            n_conditions (int): the number of conditions in the list. The array returned contains integers from 1
                to the value of `n_conditions`.
            n_reps (int): number that each element is repeated. Length of the returned array is n_conditions * n_reps.
        Returns:
            (numpy.ndarray): randomized sequence.
        """
        return numpy.random.permutation(numpy.tile(list(range(1, n_conditions+1)), n_reps))

    def get_future_trial(self, n=1):
        """
        Returns the condition of a trial n iterations into the future or past, without advancing the trials.

        Arguments:
            n (int): number of iterations into the future or past (negative numbers).
        Returns:
            (any): element of the list stored in the `conditions` attribute belonging to the trial n
                iterations into the past/future. Returns None if attempting to go beyond the first/last trial
        """
        if n > self.n_remaining or self.this_n + n < 0:
            return None
        return self.conditions[self.trials[self.this_n + n]-1]

    def transitions(self):
        """
        Count the number of transitions between conditions.

        Returns:
            (numpy.ndarray): table of shape `n_conditions` x `n_conditions` where the rows represent the condition
            transitioning from and the columns represent the condition transitioning to. For example [0, 2] shows the
            number of transitions from condition 1 to condition 3. If the `kind` of the sequence is "non_repeating",
            the diagonal is 0 because no condition transitions into itself.
        """
        transitions = numpy.zeros((self.n_conditions, self.n_conditions))
        for i, j in zip(self.trials, self.trials[1:]):
            transitions[i-1, j-1] += 1
        return transitions

    def condition_probabilities(self):
        """
        Return the frequency with which each condition appears in the sequence.

        Returns:
             (list): list of floats floats, where every element represents the frequency of one condition.
                The fist element is the frequency of the first condition and so on.
        """
        probabilities = []
        for i in range(self.n_conditions):
            num = self.trials.count(i)
            num /= self.n_trials
            probabilities.append(num)
        return probabilities

    def response_summary(self):
        """
        Generate a summary of the responses for each condition. The function counts how often a specific response
        was given to a condition for all conditions and each possible response (including None).

        Returns:
            (list of lists | None): indices of the outer list represent the conditions in the sequence. Each inner
            list contains the number of responses per response key, with the response keys sorted in ascending order,
            the last element always represents None. If the sequence is not finished yet, None is returned.
        Examples::

            import slab
            import random
            sequence = slab.Trialsequence(conditions=3, n_reps=10)  # a sequence with three conditions
            # iterate trough the list and generate a random response. The response can be either yes (1), no (0) or
            # there can be no response at all (None)
            for trial in sequence:
                response = random.choice([0, 1, None])
                sequence.add_response(response)
            sequence.response_summary()
            # Out: [[1, 1, 7], [2, 5, 3], [4, 4, 2]]
            # The first sublist shows that the subject responded to the first condition once with no (0),
            # once with yes (1) and did not give a response seven times, the second and third list show
            # prevalence of the same response keys for conditions two and three.
        """
        if self.finished:
            # list of used response key codes (add None in case it's not present):
            response_keys = [item for sublist in self.data for item in sublist]
            response_keys = list(set(response_keys + [None]))
            response_keys = sorted(response_keys, key=lambda x: (x is None, x))  # sort, with 'None' at the end
            responses = []
            for condition in self.conditions:
                idx = [i for i, cond in enumerate(self.trials) if cond == condition]  # indices of condition in sequence
                # count how often each type of key was given to this condition:
                condition_data = [self.data[i] for i in idx]
                count = collections.Counter([item for sublist in condition_data for item in sublist])
                resp_1cond = []
                for r in response_keys:
                    resp_1cond.append(count[r])
                responses.append(resp_1cond)
            return responses
        else:
            return None

    def plot(self, axis=None, show=True):
        """
        Plot the trial sequence as scatter plot.

        Arguments:
            axis (matplotlib.pyplot.Axes): plot axis to draw on, if none a new plot is generated
            show (bool): show the plot immediately, defaults to True
        """
        if plt is None:
            raise ImportError('Plotting requires matplotlib!')
        if axis is None:
            axis = plt.subplot()
        axis.scatter(range(self.n_trials), self.trials)
        axis.set(title='Trial sequence', xlabel='Trials', ylabel='Condition index')
        if show:
            plt.show()


class Staircase(collections.abc.Iterator, LoadSaveMixin, TrialPresentationOptionsMixin):
    """
    Class to handle adaptive testing which means smoothly the selecting next trial, report current values and so on.
    The sequence will terminate after a certain number of reverals have been exceeded.

    Arguments:
        start_val (int | float): initial stimulus value for the staircase
        n_reversals (int): number of reversals needed to terminate the staircase
        step_sizes (int | list): Size of steps in the staircase. Given an integer, the step size is constant. Given
            a list, the step size will progress to the next entry at each reversal. If the list is exceeded before the
            sequence was finished, it will continue with the last entry of the list as constant step size.
        step_up_factor: allows different sizes for up and down steps to implement a Kaernbach1991 weighted
            up-down method. step_sizes sets down steps, which are multiplied by step_up_factor to obtain up step
            sizes. The default is 1, i.e. same size for up and down steps.
        n_pretrials (int): number of trial at the initial stimulus value presented as before start of the staircase
        n_up (int): number of `incorrect` (or 0) responses before the staircase level increases. Is 1, regardless of
            specified value until the first reversal. Lewitt (1971) gives the up-down values for different threshold
            points on the psychometric function: 1-1 (0.5), 1-2 (0.707), 1-3 (0.794), 1-4 (0.841), 1-5 (0.891).
        n_down (int): number of `correct` (or 1) responses before the staircase level decreases (see `n_up`).
        step_type (str): defines the change of stimulus intensity at each step of the staircase. possible inputs are
            'lin' (adds or subtract a certain amount), 'db', and 'log' (prevents the intensity from reaching zero).
        min_val (int or float): smallest stimulus value permitted, or -inf for staircase without lower limit
        max_val (int or float): largest stimulus value permitted, or inf for staircase without upper limit
        label (str): text label for the sequence, defaults to an empty string
    Attributes:
        .this_trial_n: number of completed trials
        .intensities: presented stimulus values
        .current_direction: 'up' or 'down'
        .data: list of responses
        .reversal_points: indices of reversal trials
        .reversal_intensities: stimulus values at the reversals (used to compute threshold)
        .finished: True/False: have we finished yet?
    Examples::

        stairs = Staircase(start_val=50, n_reversals=10, step_type='lin',
            step_sizes=[4,2], min_val=10, max_val=60, n_up=1, n_down=1)
        print(stairs)
        for trial in stairs:
            response = stairs.simulate_response(30)
            stairs.add_response(response)
        print(f'reversals: {stairs.reversal_intensities}')
        print(f'mean of final 6 reversals: {stairs.threshold()}')
    """
    def __init__(self, start_val, n_reversals=None, step_sizes=1, step_up_factor=1, n_pretrials=0, n_up=1,
                 n_down=2, step_type='lin', min_val=-numpy.inf, max_val=numpy.inf, label=''):
        self.label = label
        self.start_val = start_val
        self.n_up = n_up
        self.n_down = n_down
        self.step_type = step_type
        try:
            self.step_sizes = list(step_sizes)
        except TypeError:
            self.step_sizes = [step_sizes]
        self._variable_step = True if len(self.step_sizes) > 1 else False
        self.step_up_factor = step_up_factor
        self.step_size_current = self.step_sizes[0]
        if n_reversals is None:
            if len(self.step_sizes) == 1:
                self.n_reversals = 8  # if Staircase called without parameters, construct a short 8-reversal test
            else:
                self.n_reversals = len(self.step_sizes) + 1  # otherwise dependent on number of step sizes
        elif len(self.step_sizes) > n_reversals:
            print(
                f'Increasing number of minimum required reversals to the number of step sizes, {len(self.step_sizes)}')
            self.n_reversals = len(self.step_sizes)
        else:
            self.n_reversals = n_reversals
        self.finished = False
        self.n_pretrials = n_pretrials
        self.this_trial_n = -n_pretrials
        self.data = []
        self.intensities = []
        self.reversal_points = []
        self.reversal_intensities = []
        self.current_direction = 'down'
        self.correct_counter = 0
        self._next_intensity = self.start_val
        self.min_val = min_val
        self.max_val = max_val
        self.pf_intensities = None  # psychometric function, auto set when finished
        self.pf_percent_correct = None  # psychometric function, auto set when finished
        self.pf_responses_per_intensity = None  # psychometric function, auto set when finished

    def __next__(self):
        """
        Is called when iterating trough a sequenceAdvances to next trial and returns it. Updates attributes
        this_trial, this_n, and this_index. If the trials have ended this method will raise a StopIteration.

        Returns:
            (int | float | StopIteration): the intensity for the next trial which is calculated by the
                `_next_intensity` method. If the sequence is finished a StopIteration is returned instead.
        """
        if not self.finished:
            self.this_trial_n += 1  # update pointer for next trial
            self.intensities.append(self._next_intensity)
            return self._next_intensity
        else:
            self._psychometric_function()  # tally responses to create a psychometric function
            raise StopIteration

    def __repr__(self):
        return self.__dict__.__repr__()

    def __str__(self):
        return f'Staircase {self.n_up}up-{self.n_down}down, trial {self.this_trial_n},' \
               f' {len(self.reversal_intensities)} reversals of {self.n_reversals}'

    def add_response(self, result, intensity=None):
        """
        Add a True or 1 to indicate a correct/detected trial
        or False or 0 to indicate an incorrect/missed trial.
        This is essential to advance the staircase to a new intensity level.
        Supplying an `intensity` value indicates that you did not use
        the recommended intensity in your last trial and the staircase will
        replace its recorded value with the one supplied.
        """
        if self._next_intensity <= self.min_val:  # always record False if at min_val
            result = False
        else:
            result = bool(result)
        self.data.append(result)
        if intensity is not None:
            self.intensities.pop()
            self.intensities.append(intensity)
        if self.this_trial_n > 0:  # we're out of the pretrials
            if result:  # correct response
                if len(self.data) > 1 and self.data[-2] == result:
                    self.correct_counter += 1  # increment if on a run
                else:
                    self.correct_counter = 1  # or reset
            else:  # incorrect response
                if len(self.data) > 1 and self.data[-2] == result:
                    self.correct_counter -= 1  # decrement if on a run
                else:
                    self.correct_counter = -1  # or reset
            self.calculate_next_intensity()

    def calculate_next_intensity(self):
        """ Based on current intensity, counter of correct responses, and current direction. """
        # TODO: description of how the current intensity is calculated
        if not self.reversal_intensities:  # no reversals yet
            if self.data[-1] is True:  # last answer correct
                reversal = bool(self.current_direction == 'up')  # got it right
                self.current_direction = 'down'
            else:  # got it wrong
                reversal = bool(self.current_direction == 'down')
                self.current_direction = 'up'
        elif self.correct_counter >= self.n_down:  # n right, time to go down!
            reversal = bool(self.current_direction != 'down')
            self.current_direction = 'down'
        elif self.correct_counter <= -self.n_up:  # n wrong, time to go up!
            reversal = bool(self.current_direction != 'up')
            self.current_direction = 'up'
        else:  # same as previous trial
            reversal = False
        if reversal:  # add reversal info
            self.reversal_points.append(self.this_trial_n)
            self.reversal_intensities.append(self.intensities[-1])
        if len(self.reversal_intensities) >= self.n_reversals:
            self.finished = True  # we're done
        # if reversal and self._variable_step:  # new step size if necessary
            # if beyond the list of step sizes, use the last one
        if len(self.reversal_intensities) >= len(self.step_sizes):
            self.step_size_current = self.step_sizes[-1]
        else:
            _sz = len(self.reversal_intensities)
            self.step_size_current = self.step_sizes[_sz]
        if self.current_direction == 'up':
            self.step_size_current *= self.step_up_factor  # apply factor for weighted up/down method
        if not self.reversal_intensities:
            if self.data[-1] == 1:
                self._intensity_dec()
            else:
                self._intensity_inc()
        elif self.correct_counter >= self.n_down:
            self._intensity_dec()  # n right, so going down
        elif self.correct_counter <= -self.n_up:
            self._intensity_inc()  # n wrong, so going up

    def _intensity_inc(self):
        """ increment the current intensity and reset counter. """
        if self.step_type == 'db':
            self._next_intensity *= 10.0**(self.step_size_current/20.0)
        elif self.step_type == 'log':
            self._next_intensity *= 10.0**self.step_size_current
        elif self.step_type == 'lin':
            self._next_intensity += self.step_size_current
        if (self.max_val is not None) and (self._next_intensity > self.max_val):
            self._next_intensity = self.max_val  # check we haven't gone out of the legal range
        self.correct_counter = 0

    def _intensity_dec(self):
        """ decrement the current intensity and reset counter. """
        if self.step_type == 'db':
            self._next_intensity /= 10.0**(self.step_size_current/20.0)
        if self.step_type == 'log':
            self._next_intensity /= 10.0**self.step_size_current
        elif self.step_type == 'lin':
            self._next_intensity -= self.step_size_current
        self.correct_counter = 0
        if (self.min_val is not None) and (self._next_intensity < self.min_val):
            self._next_intensity = self.min_val  # check we haven't gone out of the legal range

    def threshold(self, n=0):
        """
        Returns the average of the last n reversals.

        Arguments:
            n (int): number of reversals to average over, if 0 use `n_reversals` - 1.
        Returns:
            the arithmetic (if `step_type`==='lin') or geometric mean of the `reversal_intensities`.
        """
        if self.finished:
            if n == 0 or n > self.n_reversals:
                n = int(self.n_reversals) - 1
            if self.step_type == 'lin':
                return numpy.mean(self.reversal_intensities[-n:])
            return numpy.exp(numpy.mean(numpy.log(self.reversal_intensities[-n:])))
        return None  # still running the staircase

    def print_trial_info(self):
        """ Convenience method for printing current trial information. """
        print(
            f'{self.label} | trial # {self.this_trial_n}: reversals: {len(self.reversal_points)}/{self.n_reversals},'
            f' intensity {round(self.intensities[-1],2) if self.intensities else round(self._next_intensity,2)},'
            f' going {self.current_direction}, response {self.data[-1] if self.data else None}')

    def save_csv(self, filename):
        """
        Write a csv text file with the stimulus values in the 1st line and the corresponding responses in the 2nd.

        Arguments:
            filename (str): the name under which the csv file is saved.
        Returns:
            (bool): True if saving was successful, False if there are no trials to save.
        """
        if self.this_trial_n < 1:
            return False  # no trials to save
        with open(filename, 'w') as f:
            raw_intensities = str(self.intensities)
            raw_intensities = raw_intensities.replace('[', '').replace(']', '')
            f.write(raw_intensities)
            f.write('\n')
            responses = str(numpy.multiply(self.data, 1))  # convert to 0 / 1
            responses = responses.replace('[', '').replace(']', '')
            responses = responses.replace(' ', ', ')
            f.write(responses)
        return True

    def plot(self, axis=None, show=True):
        """
        Plot the staircase. If called after each trial, one plot is created and updated.

        Arguments:
            axis (matplotlib.pyplot.Axes): plot axis to draw on, if none a new plot is generated
            show (bool): whether to show the plot right after drawing.
        """
        if plt is None:
            raise ImportError('Plotting requires matplotlib!')
        if self.intensities:  # plotting only after first response
            x = numpy.arange(-self.n_pretrials, len(self.intensities)-self.n_pretrials)
            y = numpy.array(self.intensities)  # all previously played intensities
            responses = numpy.array(self.data)
            if axis is None:
                fig = plt.figure('stairs')  # figure 'stairs' is created or made current
                axis = fig.gca()
            axis.clear()
            axis.plot(x, y)
            axis.set_xlim(-self.n_pretrials, max(20, (self.this_trial_n + 15)//10*10))
            axis.set_ylim(min(0, min(y)) if self.min_val == -numpy.inf else self.min_val,
                          max(y) if self.max_val == numpy.inf else self.max_val)
            # plot green dots at correct/yes responses
            axis.scatter(x[responses], y[responses], color='green')
            # plot red dots at correct/yes responses
            axis.scatter(x[~responses], y[~responses], color='red')
            axis.scatter(len(self.intensities)-self.n_pretrials+1, self._next_intensity, color='grey')  # current trial
            axis.set_ylabel('Dependent variable')
            axis.set_xlabel('Trial')
            axis.set_title('Staircase')
            if self.finished:
                axis.hlines(self.threshold(), min(x), max(x), 'r')
            plt.draw()
            if show:
                plt.pause(0.01)

    @staticmethod
    def close_plot():
        """ Closes a staircase plot (if not drawn into a specified axis) - used for plotting after each trial. """
        plt.close('stairs')

    def _psychometric_function(self):
        """
        Create a psychometric function by binning data from a staircase procedure.
        Called automatically when staircase is finished. Sets attributes `pf_intensites` (array of intensity values
        where each is the center of an intensity bin), `pf_percent_correct` (array of mean percent correct in each bin),
        `pf_responses_per_intensity` (array of number of responses contributing to each mean).
        """
        intensities = numpy.array(self.intensities)
        responses = numpy.array(self.data)
        binned_resp = []
        binned_intensities = []
        n_points = []
        intensities = numpy.round(intensities, decimals=8)
        unique_intensities = numpy.unique(intensities)
        for this_intensity in unique_intensities:
            these_responses = responses[intensities == this_intensity]
            binned_intensities.append(this_intensity)
            binned_resp.append(numpy.mean(these_responses))
            n_points.append(len(these_responses))
        self.pf_intensities = binned_intensities
        self.pf_percent_correct = binned_resp
        self.pf_responses_per_intensity = n_points


class ResultsFile:
    """
    A class for simplifying the typical use cases of results files, including generating the name,
    creating the folders, and writing to the file after each trial. Writes a JSON Lines file,
    in which each line is a valid self-contained JSON string (see http://jsonlines.org).

    Arguments:
        subject (str): determines the name of the sub-folder and files.
        folder (None | str): folder in which all results are saved, defaults to global variable results_folder.
    Attributes:
        .path: full path to the results file.
        .subject: the subject's name.
    Example::

        ResultsFile.results_folder = 'MyResults'
        file = ResultsFile(subject='MS')
        print(file.name)
    """

    name = property(fget=lambda self: self.path.name, doc='The name of the results file.')

    def __init__(self, subject='test', folder=None, filename=None):
        self.subject = subject
        if folder is None:
            folder = results_folder
        filename = '_'.join(filter(None, (subject, filename)))
        self.path = pathlib.Path(folder / pathlib.Path(subject) / pathlib.Path(filename +
                                 datetime.datetime.now().strftime("_%Y-%m-%d-%H-%M-%S") + '.txt'))
        # make the Results folder and subject subfolder
        self.path.parent.mkdir(parents=True, exist_ok=True)

    def write(self, data, tag=None):
        """
        Safely write data to the file which is opened just before writing and closed immediately after
        to avoid data loss. Call this method at the end of each trial to save the response and trial state.

        Arguments:
            data (any): data to save must be JSON serializable [string, list, dict, ...]). If data is an
                object, the __dict__ is extracted and saved.
            tag (str): The tag is prepended as a key. If None is provided, the current time is used.
        """
        if hasattr(data, "__dict__"):
            data = data.__dict__
        try:
            data = json.loads(data)  # if payload is already json, parse it into python object
        except (json.JSONDecodeError, TypeError):
            pass  # if payload is not json - all good, will be encoded later
        if tag is None or tag == 'time':
            tag = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        with open(self.path, 'a') as file:
            file.write(json.dumps({tag: data}))
            file.write('\n')

    @staticmethod
    def read_file(filename, tag=None):
        """
        Read a results file and return the content.

        Arguments:
            filename (str | pathlib.Path):
            tag (None | str):
        Returns:
            (list | dict): The content of the file. If tag is None, the whole file is returned,
                else only the dictionaries with that tag as a key are returned. The content will
                be a list of dictionaries or a dictionary if there is only a single element.
        """
        content = []
        with open(filename) as file:
            if tag is None:
                for line in file:
                    content.append(json.loads(line))
            else:
                for line in file:
                    jwd = json.loads(line)
                    if tag in jwd:
                        content.append(jwd[tag])
        if len(content) == 1:  # if only one item in list
            content = content[0]  # get rid of the list
        return content

    def read(self, tag=None):
        """ Wrapper for the read_file method. """
        return ResultsFile.read_file(self.path, tag)

    @staticmethod
    def previous_file(subject=None):
        """
        Returns the name of the most recently used results file for a given subject.
        Intended for extracting information from a previous file when running partial experiments.

        Arguments:
            subject (str): the subject name name under which the file is stored.
        Returns:
            (pathlib.Path): full path to the most recent results file.
        """
        path = pathlib.Path(results_folder) / pathlib.Path(subject)
        files = [f for f in path.glob(subject + '*') if f.is_file()]
        files.sort()
        return files[-1]

    def clear(self):
        """ Clears the file by erasing all content. """
        with open(self.path, 'w') as file:
            file.write('')


class ResultsTable(ResultsFile):
    """
    A class for foolproof writing of results tables as comma separated values (CSV), including
    generating the name, creating the folders, and writing to the file after each trial. On
    initialization you have to provide the column headers of the CSV table, either as a list,
    a comma_separated string, or as Path object to a separate text file containing the column
    headers.
    The ResultsTable object

    Arguments:
        subject (str): determines the name of the sub-folder and files.
        columns (list | str | path.Path): list of column names or comma-separated column names
            in a string (read from a text file if Path object is given). Must be valid Python variable
            names! Will be the first row of the results file. Adding rows to the table requires giving
            a value for each variable name. (This is enforced through a namedtuple of these variables.)
        folder (None | str): folder in which all results are saved, defaults to global variable results_folder.
    Attributes:
        .path: full path to the results file.
        .subject: the subject's name.
        .name: file name
        .Row: Namedtuple with column names as fields. Must be fully populated to write to the table.
    Example::

        ResultsTable.results_folder = 'MyResults'
        header = 'timestamp, subject, trial, stimulus, response'
        # OR: header = ('timestamp', 'subject', etc.)
        # OR: header = path.Path('header_names.csv')
        table = ResultsTable(subject='MS', columns=header)
        print(table.name) # this file is now created and contains the header row
        print(table.Row) # a namedtuple has also been created with the header attributes
        print(table.Row._fields) # the fields are the column names
        # to write a row of results at the end of the trial loop:
        row = table.Row(timestamp=datetime.now(), subject=table.subject, trail=stairs.this_n, stimulus=stim.name, response=button)
        table.write(row)
    """

    def __init__(self, columns, subject='test', folder=None, filename=None):
        super().__init__(subject, folder, filename)
        self.Row = self._make_Row(columns)
        self._write_header()

    @staticmethod
    def _make_Row(columns):
        "Generate a namedtuple with fields corresponding to column names. Called automatically during init."
        if isinstance(columns, pathlib.Path):
            with open(columns) as f:
                first_line = f.readline().strip('\n')
                field_names = [x.strip() for x in first_line.split(',')]
        elif isinstance(columns, str):
            field_names = [x.strip() for x in columns.split(',')]
        else:
            field_names = columns
        return collections.namedtuple('Row', field_names)

    def write(self, row):
        """
        Safely write data to the file which is opened just before writing and closed immediately
        after to avoid data loss. Call this method at the end of each trial to save the response
        and trial state. Values with commas are enclosed in double quotes to keep the CSV valid.

        Arguments:
            row (self.Row): All values to be written have to be provided as namedtuple, the fields
                of which were generated at initialization time and cannot be changed. This ensures
                table consistency.
        Example::
            # to write a row of results at the end of the trial loop, first make a new instance of Row:
            row = table.Row(timestamp=datetime.now(), subject=table.subject, trail=stairs.this_n, stimulus=stim.name, response=button)
            # then write these row values to the file:
            table.write(row)
        """
        if not isinstance(row, self.Row):
            raise TypeError('Data has to be given as instance of namedtuple Row.')
        with open(self.path, 'a') as file:
            vals = ['\"' + str(v) + '\"' if ',' in str(v) else str(v) for v in row._asdict().values()]
            file.write(',\t'.join(vals) + '\n')

    def _write_header(self):
        "Writes the column names as header to the file, separated by commas. Called automatically during init."
        with open(self.path, 'w') as file:
            file.write(','.join(self.Row._fields) + '\n')

    def read_file(self, *args):
        raise NotImplementedError('Use pandas.read_csv or csv directly.')

    def read(self, *args):
        raise NotImplementedError('Use pandas.read_csv or csv directly.')


class Precomputed(list):
    """
    This class is a list of pre-computed sound stimuli which simplifies their generation and presentation. It is
    typically used when stimulus generation takes too long to happen in each trial. In this case, a list of stimuli is
    precomputed and a random stimulus from the list is presented in each trial, ideally without direct repetition.
    The Precomputed list has a play method which automatically selects an
    element other than the previous one for playing, and can be used like an `slab.Sound` object.

    Arguments:
        sounds (list | callable | iterator): sequence of Sound objects (each must have a play method).
        n: only used if sounds is a callable, calls it n times to make the stimuli.
    Attributes:
        .sequence: a list of all the elements that have been played already.
    Examples::

        stims = slab.Precomputed(sound_list) # using a pre-made list
        # using a lambda function to make 10 examples of pink noise
        stims = slab.Precomputed(lambda: slab.Sound.pinknoise(), n=10)
        stims = slab.Precomputed( (slab.Sound.vowel(vowel=v) for v in ['a','e','i']) ) # using a generator
        stims.play() # playing a sound from the list
    """

    def __init__(self, sounds, n=10):
        if isinstance(sounds, (list, tuple)):  # a list was passed, use as is
            list.__init__(self, sounds)
        elif callable(sounds):  # a function to generate sound objects was passed, call the function n times
            list.__init__(self, [])
            for _ in range(int(n)):
                list.append(self, sounds())
        elif isinstance(sounds, str):  # string is interpreted as name of a zip file containing the sounds
            with zipfile.ZipFile(sounds) as zipped:
                files = zipped.namelist()
                if files:
                    list.__init__(self, [])
                    for file in files:
                        list.append(self, slab.Sound(file))
        elif hasattr(sounds, '__iter__'):  # it's an iterable object, just iterate through it
            for sound in sounds:
                list.append(self, sound)
        else:
            raise TypeError('Unknown type for list argument.')
        if not all(hasattr(sound, 'play') for sound in self):
            raise TypeError('Cannot play all of the provided items.')  # all items in list need to have a play method
        self.sequence = []  # keep a list of indices of played stimuli, in case needed for later analysis

    def play(self):
        """ Play a random, but never the previous, stimulus from the list. """
        if self.sequence:
            previous = self.sequence[-1]
        else:
            previous = None
        idx = previous
        while idx == previous:
            idx = numpy.random.randint(len(self))
        self.sequence.append(idx) # add to the list of played stimuli
        self[idx].play()

    def random_choice(self, n=1):
        """
        Pick (without replacement) random elements from the list.

        Arguments:
             n (int): number of elements to pick.
        Returns:
            (list): list of n random elements.
        """
        idxs = numpy.random.randint(0, len(self), size=n)
        return [self[i] for i in idxs]

    def write(self, filename):
        """
        Save the Precomputed object as a zip file containing all sounds as wav files.

        Arguments:
            filename (str | pathlib.Path): full path to under which the file is saved.
        """
        with zipfile.ZipFile(filename, mode='a') as zip_file:  # open an empty zip file in 'append' mode
            for idx, sound in enumerate(self):
                f = io.BytesIO()  # open a virtual file (file-like memory buffer)
                sound.write(f)  # write sound to virtual file
                f.seek(0)  # rewind the file so we can read it from start
                zip_file.writestr(f's_{idx}.wav', f.read())
                f.close()

    @staticmethod
    def read(filename):
        """
        Read a zip file containing wav files.

        Arguments:
            filename (str | pathlib.Path): full path to the file to be read.
        Returns:
            (slab.Precomputed): the file content.
        """

        stims = Precomputed([])
        with zipfile.ZipFile(filename, 'r') as zipped:
            files = zipped.namelist()
            for file in files:
                wav_bytes = zipped.read(file)
                stims.append(slab.Sound.read(io.BytesIO(wav_bytes)))
        return stims


def load_config(filename):
    """
    Reads a text file with variable assignments. This is a simple convenience method that allows easy writing and
    loading of configuration text files. Experiments sometimes use configuration files when experimenters (who might
    not by Python programmers) need to set parameters without changing the code. The format is a plain text file with a
    variable assignment on each line, because it is meant to be written and changed by humans. These variables and their
    values are then accessible as a namedtuple.

    Arguments:
        filename (str | pathlib.Path): path to the file to be read.
    Returns:
        (collections.namedtuple): a tuple containing the variables and values defined in the text file.
    Example::

        # assuming there is a file named 'example.txt' with the following content:
        samplerate = 32000
        pause_duration = 30
        speeds = [60,120,180]
        # call load_config to parse the file into a named tuple:
        conf = load_config('example.txt')
        conf.speeds
        # Out: [60, 120, 180]
    """
    with open(filename, 'r') as f:
        lines = f.readlines()
    if lines:
        var_names = []
        values = []
        for line in lines:
            var, val = line.strip().split('=')
            var_names.append(var.strip())
            values.append(eval(val.strip()))
        config_tuple = collections.namedtuple('config', var_names)
        return config_tuple(*values)
